# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import hdtv.color

from hdtv.weakref_proxy import weakref
from hdtv.spectrum import CutSpectrum
from hdtv.drawable import DrawableManager


class Matrix(DrawableManager):
    def __init__(self, histo2D, sym, viewport):
        DrawableManager.__init__(self, viewport)
        self.histo2D = histo2D
        self.sym = sym
        self.ID = None
        self._xproj = None
        self._yproj = None
        self._color = hdtv.color.default

    # color property
    def _set_color(self, color):
        # give all cuts and projections the same color
        self._color = color
        for cut in self.dict.values():
            cut.color = color
        if self._xproj is not None:
            self._xproj.color = color
        if self._yproj is not None:
            self._yproj.color = color

    def _get_color(self):
        return self._color

    color = property(_get_color, _set_color)

    # name
    @property
    def name(self):
        return self.histo2D.name

    # projections
    @property
    def xproj(self):
        return self.project(axis="x")

    @property
    def yproj(self):
        return self.project(axis="y")

    def project(self, axis):
        """
        Return Projection along a given axis
        """
        if getattr(self, "_%sproj" % axis) is None:
            # no valid link to that projection, create fresh object
            hist = getattr(self.histo2D, "%sproj" % axis)
            if self.sym:
                a = "0"
            else:
                a = axis
            proj = CutSpectrum(hist, self, axis=a)
            proj.color = self.color
            setattr(self, "_%sproj" % axis, weakref(proj))
            return proj
        else:
            return getattr(self, "_%sproj" % axis)

    def ExecuteCut(self, cut):
        cutHisto = self.histo2D.ExecuteCut(cut.regionMarkers, cut.bgMarkers, cut.axis)
        if cut.axis == "x":
            axis = "y"
        elif cut.axis == "y":
            axis = "x"
        elif self.sym:
            axis = "0"
        else:
            raise RuntimeError
        cutSpec = CutSpectrum(cutHisto, self, axis)
        cutSpec.color = self.color
        return cutSpec

    # overwrite some functions from Drawable
    def Insert(self, obj, ID=None):
        """
        Insert cut to internal dict
        """
        obj.color = self.color
        obj.dashed = True
        if ID is None:
            ID = self.activeID
        ID = DrawableManager.Insert(self, obj, ID)
        self.ActivateObject(ID)
        return ID

    def ActivateObject(self, ID):
        if ID is not None and ID not in self.ids:
            raise KeyError
        # housekeeping for old active cut
        if self.activeID is not None:
            cut = self.GetActiveObject()
            cut.active = False
        # change active ID
        self.activeID = ID
        self._iteratorID = self.activeID
        # housekeeping for new active cut
        if self.activeID is not None:
            cut = self.GetActiveObject()
            cut.active = True

    def Show(self, axis="0"):
        for ID in self.visible:
            if self.dict[ID].axis == axis:
                self.dict[ID].Show()



#!/usr/bin/env python3

# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2019  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
This is the main HDTV application.
"""

import sys
import os
import glob
from pathlib import Path
import argparse

if __name__ == "__main__":
    project_dir = Path(__file__).absolute().parents[1].resolve()
    project_path = str((project_dir / "hdtv").resolve())
    if project_path in sys.path:
        sys.path.remove(project_path)
    sys.path.insert(0, str(project_dir))


def check_root_version():
    import ROOT

    version = ROOT.gROOT.GetVersion()

    if version in ["6.22/00", "6.22/02"]:
        print(
            f"Error: Unsupported ROOT version {version} detected.\n"
            "ROOT 6.22/00 and 6.22/02 are not compatible with HDTV.\n"
            "Please upgrade to ROOT 6.22/04 or later "
            "or use a different major version.\n See README.md for more information.",
        )
        exit(1)


class App:
    def __init__(self):
        # Reset command line arguments so that ROOT does not stumble about them
        hdtv_args = sys.argv[1:]
        sys.argv = [sys.argv[0]]

        # Get config and data directory
        self.legacypath = Path.home() / ".hdtv"
        self.configpath = Path(
            os.getenv(
                "HDTV_USER_PATH",
                self.legacypath
                if self.legacypath.is_dir()
                else Path(os.getenv("XDG_CONFIG_HOME", Path.home() / ".config"))
                / "hdtv",
            )
        )
        self.datapath = Path(
            os.getenv(
                "HDTV_USER_PATH",
                self.legacypath
                if self.legacypath.is_dir()
                else Path(os.getenv("XDG_DATA_HOME", Path.home() / ".local/share"))
                / "hdtv",
            )
        )

        for path in [self.datapath, self.configpath]:
            try:
                path.mkdir(parents=True)
            except OSError:
                pass

        if not os.access(self.datapath, os.W_OK):
            print(
                f"ERROR: data path {self.datapath} is not writable",
                file=sys.stderr,
            )

        if not os.access(self.configpath, os.R_OK):
            print(
                "ERROR: Could not access config path " + self.configpath,
                file=sys.stderr,
            )

        os.environ["HDTV_USER_PATH"] = str(self.configpath)
        sys.path.append(str(self.configpath))
        sys.path.append(str(self.configpath / "plugins"))

        args = self.parse_args(hdtv_args)

        if args.rebuildusr is not None:
            import hdtv.rootext.dlmgr

            hdtv.rootext.dlmgr.RebuildLibraries(
                hdtv.rootext.dlmgr.usrdir, libraries=args.rebuildusr or None
            )
        if args.rebuildsys is not None:
            import hdtv.rootext.dlmgr

            hdtv.rootext.dlmgr.RebuildLibraries(
                hdtv.rootext.dlmgr.sysdir, libraries=args.rebuildsys or None
            )

        if args.rebuildusr or args.rebuildsys:
            sys.exit(0)

        check_root_version()

        # Import core modules
        import hdtv.cmdline
        import hdtv.session
        import hdtv.ui

        hdtv.cmdline.SetHistory(self.datapath / "hdtv_history")
        hdtv.cmdline.SetInteractiveDict(locals())
        spectra = hdtv.session.Session()
        import __main__

        __main__.spectra = spectra

        # Import core plugins
        import hdtv.plugins.textInterface
        import hdtv.plugins.ls
        import hdtv.plugins.run
        import hdtv.plugins.specInterface
        import hdtv.plugins.fitInterface
        import hdtv.plugins.calInterface
        import hdtv.plugins.matInterface
        import hdtv.plugins.rootInterface
        import hdtv.plugins.config
        import hdtv.plugins.fitlist
        import hdtv.plugins.fittex
        import hdtv.plugins.fitmap
        import hdtv.plugins.dblookup
        import hdtv.plugins.peakfinder
        import hdtv.plugins.printing

        hdtv.ui.msg("HDTV - Nuclear Spectrum Analysis Tool")

        # Execute startup.py for user configuration in python
        try:
            import startup
        except ImportError:
            hdtv.ui.debug("No startup.py file")

        # Execute startup.hdtv and startup.hdtv.d/*.hdtv
        # for user configuration in "hdtv" language
        startup_d_hdtv = [self.configpath / "startup.hdtv"] + list(
            (self.configpath / "startup.hdtv.d").glob("*.hdtv")
        )

        for startup_hdtv in startup_d_hdtv:
            try:
                if os.path.exists(startup_hdtv):
                    hdtv.cmdline.command_line.ExecCmdfile(startup_hdtv)
            except OSError as msg:
                hdtv.ui.error(f"Error reading {startup_hdtv}: {msg}")

        self.run_batchfile(args)
        self.run_commands(args)

        hdtv.cmdline.MainLoop()
        hdtv.plugins.rootInterface.r.rootfile = None
        hdtv.cmdline.command_tree.SetDefaultLevel(1)

    def parse_args(self, args):
        from hdtv._version import get_versions

        __version__ = get_versions()["version"]
        del get_versions
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "-b", "--batch", dest="batchfile", help="Open and execute HDTV batchfile"
        )
        parser.add_argument(
            "-e", "--execute", dest="commands", help="Execute HDTV command(s)"
        )
        parser.add_argument(
            "-v",
            "--version",
            action="version",
            help="Show HDTV Version",
            version=f"HDTV {__version__}",
        )
        parser.add_argument(
            "--rebuild-usr",
            nargs="*",
            dest="rebuildusr",
            help="Rebuild ROOT-loadable libraries for the current user",
        )
        parser.add_argument(
            "--rebuild-sys",
            nargs="*",
            dest="rebuildsys",
            help="Rebuild ROOT-loadable libraries for all users",
        )
        return parser.parse_args(args)

    def run_commands(self, args):
        """Execute commands given on command line"""
        import hdtv.cmdline

        if args.commands is not None:
            hdtv.cmdline.command_line.DoLine(args.commands)

    def run_batchfile(self, args):
        """Execute batchfile given on command line"""
        import hdtv.cmdline
        import hdtv.ui

        try:
            if args.batchfile is not None:
                hdtv.cmdline.command_line.ExecCmdfile(args.batchfile)
        except OSError as msg:
            hdtv.ui.msg(f"Error reading {args.batchfile}: {msg}")


def run():
    App()


if __name__ == "__main__":
    run()



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import pytest

from tests.helpers.utils import redirect_stdout, hdtvcmd

from hdtv.util import monkey_patch_ui

monkey_patch_ui()

import hdtv.cmdline
import hdtv.options
import hdtv.session

import __main__

try:
    if not hasattr(__main__, "spectra"):
        __main__.spectra = hdtv.session.Session()
except RuntimeError:
    pass

from hdtv.plugins.specInterface import spec_interface
import hdtv.plugins.matInterface

spectra = __main__.spectra


@pytest.fixture(autouse=True)
def prepare():
    hdtv.options.Set("table", "classic")
    hdtv.options.Set("uncertainties", "short")
    spectra.Clear()
    yield
    spectra.Clear()


@pytest.mark.skip(reason="need example matrix")
def test_cmd_matrix_get_sym(matrix):
    raise NotImplementedError


@pytest.mark.skip(reason="need example matrix")
def test_cmd_matrix_get_asym(matrix):
    raise NotImplementedError


@pytest.mark.skip(reason="need example matrix")
def test_cmd_matrix_list(matrix):
    raise NotImplementedError


@pytest.mark.skip(reason="need example matrix; hard to test")
def test_cmd_matrix_view(matrix):
    raise NotImplementedError


@pytest.mark.skip(reason="need example matrix")
def test_cmd_cut_marker(matrix):
    raise NotImplementedError


@pytest.mark.skip(reason="need example matrix")
def test_cmd_cut_execute(matrix):
    raise NotImplementedError


@pytest.mark.skip(reason="need example matrix")
def test_cmd_cut_activate(matrix):
    raise NotImplementedError


@pytest.mark.skip(reason="need example matrix")
def test_cmd_cut_clear(matrix):
    raise NotImplementedError


@pytest.mark.skip(reason="need example matrix")
def test_cmd_cut_store(matrix):
    raise NotImplementedError


@pytest.mark.skip(reason="need example matrix")
def test_cmd_cut_delete(matrix):
    raise NotImplementedError



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2019  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
Test the different implemented background models with an artificial spectrum.
The test spectrum is a sequence of step functions with increasing step height. On each step, a normal-distributed peak is located. No artificial fluctuations are introduced to give simple verifiable results.
Specific background models (for example the exponential background) may require particular shapes of the background, so their steps may be superimposed by another function.
"""

import os

import pytest
import xml.etree.ElementTree as ET

from tests.helpers.utils import hdtvcmd, isclose
from tests.helpers.fixtures import temp_file
from tests.helpers.create_test_spectrum import ArtificialSpec, ArtificialSpecProp

from numpy import exp, log, sqrt

from hdtv.util import monkey_patch_ui

monkey_patch_ui()

import hdtv.cmdline
import hdtv.options
import hdtv.session

import __main__

try:
    __main__.spectra = hdtv.session.Session()
except RuntimeError:
    pass

from hdtv.plugins.specInterface import spec_interface
from hdtv.plugins.fitInterface import fit_interface
from hdtv.plugins.fitlist import fitxml

spectra = __main__.spectra

BG_MARKER_TOLERANCE = 0.5
SIGMA_TO_FWHM = 2.3548299450  # = 2*sqrt(2*ln(2))
N_SIGMA = (
    2  # This number determines how many standard deviations (sigma) a fitted quantity
)
# may deviate from the ideal value. Sigma is either provided by some fitting algorithm or by
# statistical considerations here.
# For the former case, this test is self-consistent and assumes that the algorithm estimates the
# uncertainty correctly.
WRITE_BATCHFILE = False  # Determines whether the batch files of the test fits should be written to file.

# Initialize the test spectrum properties which are needed for the parameterized tests
ts_prop = ArtificialSpecProp()


@pytest.fixture
def test_spectrum(tmp_path):
    ts = ArtificialSpec(path=tmp_path)
    ts.create()
    return ts


@pytest.mark.parametrize(
    # nparams is the true number of parameters. The length n of the list bgparams
    # only determines that the first n parameters will be checked.
    # This is useful, for example, when a cubic or even quartic model is fit
    # to a constant background. Since Chi2 of the fit depends very strongly on the
    # values of the higher-order parameters (and even oscillates: third order cancels second order, ...),
    # the algorithm will terminate at some arbitrary
    # low value, and give a false estimate of the uncertainty.
    "model, nparams, step, nregions, settings, errormessage, bgparams, bg_volume",
    [
        #        ("polynomial", 2, 0, 3, "", "", [10., 0.], 0.),
        #        ("polynomial", 2, 0, 3, "fit parameter background 2", "", [10., 0.], 0.),
        (
            "polynomial",
            3,
            0,
            3,
            "fit parameter background 3",
            "",
            [10.0, 0.0, 0.0],
            0.0,
        ),
        ("polynomial", 3, 0, 3, "fit parameter background free", "", [10.0, 0.0], 0.0),
        ("polynomial", 4, 0, 4, "fit parameter background free", "", [10.0, 0.0], 0.0),
        ("interpolation", 6, 0, 2, "", "Background fit failed.", [], 0.0),
        (
            "interpolation",
            6,
            0,
            3,
            "",
            "",
            [
                0.5
                * (ts_prop.bg_regions[0][0] + ts_prop.bg_regions[0][1])
                * ts_prop.nbins_per_step,
                ts_prop.bg_per_bin,
                0.5
                * (ts_prop.bg_regions[1][0] + ts_prop.bg_regions[1][1])
                * ts_prop.nbins_per_step,
                ts_prop.bg_per_bin,
                0.5
                * (ts_prop.bg_regions[2][0] + ts_prop.bg_regions[2][1])
                * ts_prop.nbins_per_step,
                ts_prop.bg_per_bin,
            ],
            0.0,
        ),
        (
            "exponential",
            2,
            1,
            3,
            "fit parameter background 2",
            "",
            [log(ts_prop.bg_per_bin) + 2.0, -1.0 / (0.5 * ts_prop.nbins_per_step)],
            0.5
            * ts_prop.nbins_per_step
            * exp(log(ts_prop.bg_per_bin) + 2.0)
            * (
                exp(
                    -(
                        1.5 * ts_prop.nbins_per_step
                        - 3.0 * ts_prop.peak_width * ts_prop.nbins_per_step
                    )
                    / (0.5 * ts_prop.nbins_per_step)
                )
                - exp(
                    -(
                        1.5 * ts_prop.nbins_per_step
                        + 3.0 * ts_prop.peak_width * ts_prop.nbins_per_step
                    )
                    / (0.5 * ts_prop.nbins_per_step)
                )
            ),
        ),
    ],
)
def test_backgroundRegions(
    model,
    nparams,
    step,
    nregions,
    settings,
    errormessage,
    bgparams,
    bg_volume,
    temp_file,
    test_spectrum,
):

    spec_interface.LoadSpectra(test_spectrum.filename)
    command = ["fit function background activate " + model]
    if settings != "":
        command.append(settings)
    for i in range(nregions):
        command.append(
            "fit marker background set %f"
            % ((step + test_spectrum.bg_regions[i][0]) * test_spectrum.nbins_per_step)
        )
        command.append(
            "fit marker background set %f"
            % ((step + test_spectrum.bg_regions[i][1]) * test_spectrum.nbins_per_step)
        )
    command.append(
        "fit marker region set %f"
        % (
            (step + 0.5) * test_spectrum.nbins_per_step
            - 3.0 * test_spectrum.peak_width * test_spectrum.nbins_per_step
        )
    )
    command.append(
        "fit marker region set %f"
        % (
            (step + 0.5) * test_spectrum.nbins_per_step
            + 3.0 * test_spectrum.peak_width * test_spectrum.nbins_per_step
        )
    )
    command.append(
        "fit marker peak set %f" % ((step + 0.5) * test_spectrum.nbins_per_step)
    )

    command.append("fit execute")
    command.append("fit store")
    for i in range(nregions):
        command.append("fit marker background delete %i" % (i))
    command.append("fit marker region delete 0")
    command.append("fit marker peak delete 0")
    f, ferr = hdtvcmd(*command)

    if WRITE_BATCHFILE:
        batchfile = os.path.join(
            os.path.curdir,
            "test",
            "share",
            model + "_" + str(nparams) + "_" + str(step) + "_background.hdtv",
        )
        bfile = open(batchfile, "w")
        bfile.write("spectrum get " + test_spectrum.filename + "\n")
        for c in command:
            bfile.write(c + "\n")
        bfile.close()

    if errormessage != "":
        hdtvcmd("fit delete 0", "spectrum delete 0")
        assert errormessage in ferr
    else:
        assert ferr == ""

        fitxml.WriteXML(spectra.Get("0").ID, temp_file)
        fitxml.ReadXML(spectra.Get("0").ID, temp_file, refit=True, interactive=False)
        hdtvcmd("fit delete 0", "spectrum delete 0")

        # Parse xml file manually and check correct output
        tree = ET.parse(temp_file)
        root = tree.getroot()

        # Check the number of fits
        fits = root.findall("fit")
        assert len(fits) == 1

        # Check the number and positions of background markers
        bgMarkers = fits[0].findall("bgMarker")
        assert len(bgMarkers) == nregions
        for i in range(len(bgMarkers)):
            assert isclose(
                float(bgMarkers[i].find("begin").find("cal").text),
                (step + test_spectrum.bg_regions[i][0]) * test_spectrum.nbins_per_step,
                abs_tol=BG_MARKER_TOLERANCE,
            )
            assert isclose(
                float(bgMarkers[i].find("end").find("cal").text),
                (step + test_spectrum.bg_regions[i][1]) * test_spectrum.nbins_per_step,
                abs_tol=BG_MARKER_TOLERANCE,
            )
        # Check the number (and values) of background parameters
        background = fits[0].find("background")
        assert background.get("backgroundModel") == model
        assert int(background.get("nparams")) == nparams

        # Check the background parameters
        if len(bgparams) > 0:
            params = background.findall("param")
            for i in range(len(params)):
                if i < len(bgparams):
                    assert isclose(
                        float(params[i].find("value").text),
                        bgparams[i],
                        abs_tol=N_SIGMA * float(params[i].find("error").text),
                    )

        # Check the fit result
        # All results will be compared
        assert isclose(
            float(fits[0].find("peak").find("cal").find("vol").find("value").text),
            test_spectrum.peak_volume,
            abs_tol=N_SIGMA
            * float(fits[0].find("peak").find("cal").find("vol").find("error").text),
        )
        assert isclose(
            float(fits[0].find("peak").find("cal").find("width").find("value").text),
            SIGMA_TO_FWHM * test_spectrum.peak_width * test_spectrum.nbins_per_step,
            abs_tol=N_SIGMA
            * float(fits[0].find("peak").find("cal").find("width").find("error").text),
        )

        # Check the integration result
        integrals = fits[0].findall("integral")
        assert len(integrals) == 3

        assert integrals[0].get("integraltype") == "tot"
        assert integrals[1].get("integraltype") == "bg"
        assert integrals[2].get("integraltype") == "sub"
        # Peak volume
        assert isclose(
            float(integrals[2].find("uncal").find("vol").find("value").text),
            test_spectrum.peak_volume,
            abs_tol=N_SIGMA
            * float(integrals[2].find("uncal").find("vol").find("error").text),
        )
        # Background volume
        if bg_volume > 0.0:
            assert isclose(
                float(integrals[1].find("uncal").find("vol").find("value").text),
                bg_volume,
                abs_tol=sqrt(bg_volume),
            )
        else:
            assert isclose(
                float(integrals[1].find("uncal").find("vol").find("value").text),
                test_spectrum.bg_per_bin
                * 6.0
                * test_spectrum.peak_width
                * test_spectrum.nbins_per_step,
                abs_tol=N_SIGMA
                * sqrt(
                    test_spectrum.bg_per_bin
                    * 6.0
                    * test_spectrum.peak_width
                    * test_spectrum.nbins_per_step
                ),
            )



import os
import random
import time

# Cut using TV


class Cut:
    def __init__(self):
        self.fBgRegions = []

    def Cut(self, matfile, r1, r2):
        tempdir = "/home/braun/Diplom/temp"
        fname = "%032X.asc" % random.randint(0, 2**128)

        tvcmds = "cut activate 0; "
        tvcmds += "cut attach matrix 1; "
        tvcmds += "cut attach dir 2; "

        tvcmds += "cut dir open 2 %s; " % tempdir
        tvcmds += "cut matrix open 1 %s; " % matfile

        # Sub-channel resolution is really useless here,
        # because TV cuts only with channel resolution anyway

        tvcmds += "cut marker cut enter %.1f; " % r1
        tvcmds += "cut marker cut enter %.1f; " % r2

        for bg in self.fBgRegions:
            tvcmds += "cut marker bg-gate enter %.1f;" % bg

        tvcmds += "cut create cut; "

        tvcmds += "spec write %s'txt active; " % (tempdir + "/" + fname)

        tvcmds += "exit; "

        os.spawnl(
            os.P_WAIT,
            "/usr/bin/xvfb-run",
            "/usr/bin/xvfb-run",
            "-w",
            "0",
            "/ikp/bin/tv",
            "-src",
            "-rc",
            "-e",
            tvcmds,
        )

        time.sleep(0.5)

        return tempdir + "/" + fname



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import os
import tempfile

import pytest

from tests.helpers.utils import redirect_stdout, hdtvcmd

from hdtv.util import monkey_patch_ui

monkey_patch_ui()

import hdtv.cmdline
import hdtv.options
import hdtv.ui
import hdtv.session


import __main__

try:
    __main__.spectra = hdtv.session.Session()
except RuntimeError:
    pass

import hdtv.plugins.printing


@pytest.fixture(autouse=True)
def prepare():
    hdtv.options.Set("ui.out.level", "3")


def test_cmd_printing():
    try:
        outfile = tempfile.mkstemp(".svg", "hdtv_ptest_")[1]
        f, ferr = hdtvcmd(
            f"print -y 'my_ylabel' -x 'my_xlabel' -t 'my_title' -F {outfile}"
        )
        print(f)
        print(ferr)
        assert f == ""
        assert ferr == ""
        with open(outfile) as fout:
            result = fout.read()
        assert "my_ylabel" in result
        assert "my_xlabel" in result
        assert "my_title" in result
    finally:
        os.remove(outfile)


@pytest.mark.parametrize("fmt", ["eps", "pdf", "png", "ps", "raw", "rgba", "svg"])
# "svgz", "jpeg", "tiff" <- fail for no good reason
def test_cmd_printing_formats(fmt):
    try:
        outfile = tempfile.mkstemp("." + fmt, "hdtv_pftest_")[1]
        f, ferr = hdtvcmd(
            f"print -y 'my_ylabel' -x 'my_xlabel' -t 'my_title' -F {outfile}"
        )
        print(f)
        print(ferr)
        assert f == ""
        assert ferr == ""
    finally:
        os.remove(outfile)



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import array
import os
import ROOT
import hdtv.ui
import hdtv.rootext.mfile


class SpecReaderError(Exception):
    pass


class TextSpecReader:
    """
    Configurable formatted text file import
    The format string specifies the meaning of the columns in the file.
     x: x  (position)
     y: y  (counts)
     e: dy (error of counts)
     i: ignore this column

    If no column for x is specified, the bin number (beginning at zero) is
    taken for the bin center (just as the old tv program does it). If no
    column for e is specified, square root errors are assumed. It is an error
    if no column for y is specified.

    If no format string is given, the format is inferred from the number of
    columns in the file:
     1: y
     2: xy
     3: xye

    The cmts parameter is a tuple of possible strings that start a comment.
    cmts=None switches comment processing off. Comments extend to the end of the
    line. If unspecified, cmt defaults to ("#", "!", "//").
    """

    def __init__(self, fmt=None, cmts=("#", "!", "//")):
        self.xcol = self.ycol = self.ecol = None
        self.ncols = None
        self.cmts = cmts

        # Parse the format specifier string, if one has been specified
        if fmt:
            self.ncols = len(fmt)

            for col in range(0, len(fmt)):
                c = fmt[col]

                if c in ("x", "y", "e"):
                    if self.__dict__[c + "col"] is None:
                        self.__dict__[c + "col"] = col
                    else:
                        raise SpecReaderError(
                            "Invalid format string: %s appears more than once" % c
                        )
                elif c == "i":
                    pass
                else:
                    raise SpecReaderError("Invalid character %s in format string" % c)

            if self.ycol is None:
                raise SpecReaderError("You must specify a column for y")

    def GetBinLowEdges(self, centers):
        """
        Generate an array of (n+1) bin lower edges from an array of n bin
        centers. The result is returned as an array object so that is can be
        passed directly to the ROOT.TH1 constructor.
        """
        # This function generates n+1 lower bin edges l_0,...,l_n from n bin
        # centers c_0,...,c_{n-1}, such that the n equations
        #  c_i = (l_i + l_{i+1})/2 , i = 0,...,n-1
        # are fulfilled. Note that the problem is underdefined (n equations for
        # n+1 unknowns), so that there is a somewhat arbitrary choice being
        # made.
        xbins = array.array("d")
        w = (centers[1] - centers[0]) / 2.0
        xbins.append(centers[0] - w)
        for i in range(1, len(centers)):
            w = centers[i] - centers[i - 1] - w
            xbins.append(centers[i] - w)
        xbins.append(centers[-1] + w)

        return xbins

    def StripComments(self, line):
        end = len(line)

        if self.cmts:
            for cmt in self.cmts:
                pos = line.find(cmt)
                if pos >= 0:
                    end = min(end, pos)

        return line[:end]

    def GetSpectrum(self, fname, histname, histtitle):
        """
        Process a text file into a ROOT histogram object, using the format
        specified in the constructor.
        """
        data = []
        f = open(fname)
        linenum = 1
        x = y = e = None

        try:
            for line in f:
                # Strip comments
                line = self.StripComments(line)

                # Split line into columns
                cols = line.split()

                # Ignore empty lines
                if len(cols) == 0:
                    linenum += 1
                    continue

                # If the format string was set to autodetect, we use the number
                # of columns in the first non-empty line to determine the
                # format.
                if self.ncols is None:
                    self.ncols = len(cols)
                    if self.ncols == 1:
                        self.xcol = None
                        self.ycol = 0
                        self.ecol = None
                    elif self.ncols == 2:
                        self.xcol = 0
                        self.ycol = 1
                        self.ecol = None
                    elif self.ncols == 3:
                        self.xcol = 0
                        self.ycol = 1
                        self.ecol = 2
                    else:
                        raise SpecReaderError(
                            "%s: %d: Failed to autodetect file format: found %d columns"
                            % (fname, linenum, self.ncols)
                        )

                # Check if number of columns is consistent
                elif len(cols) != self.ncols:
                    raise SpecReaderError(
                        "%s: %d: Invalid number of columns (found=%d, expected=%d)"
                        % (fname, linenum, len(cols), self.ncols)
                    )

                # Parse specified columns into float values
                linedata = []
                for col in (self.xcol, self.ycol, self.ecol):
                    if col is not None:
                        try:
                            linedata.append(float(cols[col]))
                        except ValueError:
                            raise SpecReaderError(
                                '%s: %d: Failed to parse value "%s" into float'
                                % (fname, linenum, cols[col])
                            )
                    else:
                        linedata.append(None)

                data.append(linedata)

                linenum += 1

        finally:
            f.close()

        nbins = len(data)

        if self.xcol is not None:
            # Sort by increasing x value
            data.sort(key=lambda a: a[0])

            xbins = self.GetBinLowEdges([d[0] for d in data])
            hist = ROOT.TH1D(histname, histtitle, nbins, xbins)
        else:
            hist = ROOT.TH1D(histname, histtitle, nbins, -0.5, nbins - 0.5)

        # Fill ROOT histogram object
        for b in range(0, nbins):
            hist.SetBinContent(b + 1, data[b][1])
            if self.ecol is not None:
                hist.SetBinError(b + 1, data[b][2])

        return hist


class SpecReader:
    @staticmethod
    def GetSpectrum(fname, fmt=None, histname=None, histtitle=None):
        """
        Reads a histogram from a non-ROOT file. fmt specifies the format.
        The following formats are recognized:
          * cracow  (Cracow from GSI)
          * mfile   (use libmfile and attempt autodetection)
          * any format specifier understood by libmfile
        """
        default_format = "mfile"

        if not fmt:
            fmt = default_format
        if histname is None:
            histname = os.path.basename(fname)
        if histtitle is None:
            histtitle = os.path.basename(fname)

        if fmt.lower() == "cracow":
            # hdtv.dlmgr.LoadLibrary("cracowio")
            # cio = ROOT.CracowIO()
            # hist = cio.GetCracowSpectrum(fname, histname, histtitle)
            # if not hist:
            #     raise SpecReaderError(cio.GetErrorMsg())
            # return hist
            raise SpecReaderError("Format not longer supported")
        elif fmt.split(":")[0].lower() == "col":
            # Extract subformat specifier to pass on to TextSpecReader
            pos = fmt.find(":")
            if pos > 0:
                subfmt = fmt[pos + 1 :]
            else:
                subfmt = None

            # The following lines may raise a SpecReaderError exception
            txtio = TextSpecReader(subfmt)
            return txtio.GetSpectrum(fname, histname, histtitle)
        else:
            mhist = ROOT.MFileHist()

            if not fmt or fmt.lower() == "mfile":
                result = mhist.Open(fname)
            else:
                result = mhist.Open(fname, fmt)

            if result != ROOT.MFileHist.ERR_SUCCESS:
                raise SpecReaderError(mhist.GetErrorMsg())

            # Get spectra out of Mfile Mat structure
            # Use the spectra given by line
            level = 0
            line = 0
            if len(fmt.split(".")) == 3:
                line = int(fmt.split(".")[0]) - 1
                level = 0
                hdtv.ui.msg("Using spectra in line " + str(line))

            hist = mhist.ToTH1D(histname, histtitle, level, line)
            if not hist:
                raise SpecReaderError(mhist.GetErrorMsg())
            return hist

    @staticmethod
    def GetMatrix(fname, fmt=None, histname=None, histtitle=None):
        if histname is None:
            histname = os.path.basename(fname)
        if histtitle is None:
            histtitle = os.path.basename(fname)

        mhist = ROOT.MFileHist()

        # FIXME: error handling
        if not fmt or fmt.lower() == "mfile":
            mhist.Open(fname)
        else:
            mhist.Open(fname, fmt)

        # FIXME: this ignores possibly specified bin errors
        hist = mhist.ToTH2D(histname, histtitle, 0)
        if not hist:
            raise SpecReaderError(mhist.GetErrorMsg())
        return hist

    @staticmethod
    def GetVMatrix(fname, fmt=None, histname=None, histtitle=None):
        """
        Load a ``virtual'' matrix, i.e. a matrix that is not completely loaded
        into memory.
        """
        if histname is None:
            histname = os.path.basename(fname)
        if histtitle is None:
            histtitle = os.path.basename(fname)

        mhist = ROOT.MFileHist()

        # FIXME: error handling
        if not fmt or fmt.lower() == "mfile":
            mhist.Open(fname)
        else:
            mhist.Open(fname, fmt)

        # FIXME: this ignores possibly specified bin errors
        return ROOT.MFMatrix(mhist, 0)

    @staticmethod
    def WriteSpectrum(hist, fname, fmt):
        result = ROOT.MFileHist.WriteTH1(hist, fname, fmt)
        if result != ROOT.MFileHist.ERR_SUCCESS:
            raise SpecReaderError(ROOT.MFileHist.GetErrorMsg(result))



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
Peak Model for electron-electron scattering
Implementation requested by Oleksiy Burda <burda@ikp.tu-darmstadt.de>
"""

from uncertainties import ufloat

import ROOT
import hdtv.rootext.display
import hdtv.rootext.fit

from .peak import PeakModel
from hdtv.drawable import Drawable


class EEPeak(Drawable):
    """
    Peak object for the ee fitter
    """

    def __init__(self, pos, amp, sigma1, sigma2, eta, gamma, vol, color=None, cal=None):
        super().__init__(color, cal)
        self.pos = pos
        self.amp = amp
        self.sigma1 = sigma1
        self.sigma2 = sigma2
        self.eta = eta
        self.gamma = gamma
        # vol is not a fit parameter, but rather the result of the fit
        self.vol = vol
        # dictionary for storing additional user supplied values
        self.extras = dict()

    def __getattr__(self, name):
        """
        calculate calibrated values on the fly for pos, sigma1 and sigma2
        """
        if name == "pos_cal":
            if self.cal is None:
                return self.pos
            pos_uncal = self.pos.nominal_value
            pos_err_uncal = self.pos.std_dev
            pos_cal = self.cal.Ch2E(pos_uncal)
            pos_err_cal = abs(self.cal.dEdCh(pos_uncal) * pos_err_uncal)
            return ufloat(pos_cal, pos_err_cal)
        elif name == "sigma1_cal":
            if self.cal is None:
                return self.sigma1
            pos_uncal = self.pos.nominal_value
            sigma1_uncal = self.sigma1.nominal_value
            sigma1_err_uncal = self.sigma1.std_dev
            sigma1_cal = self.cal.Ch2E(pos_uncal) - self.cal.Ch2E(
                pos_uncal - sigma1_uncal
            )
            # This is only an approximation, valid as d(fwhm_cal)/d(pos_uncal) \approx 0
            #  (which is true for Ch2E \approx linear)
            sigma1_err_cal = abs(
                self.cal.dEdCh(pos_uncal - sigma1_uncal) * sigma1_err_uncal
            )
            return ufloat(sigma1_cal, sigma1_err_cal)
        elif name == "sigma2_cal":
            if self.cal is None:
                return self.sigma2
            pos_uncal = self.pos.nominal_value
            sigma2_uncal = self.sigma2.nominal_value
            sigma2_err_uncal = self.sigma2.std_dev
            sigma2_cal = self.cal.Ch2E(pos_uncal) - self.cal.Ch2E(
                pos_uncal - sigma2_uncal
            )
            # This is only an approximation, valid as d(fwhm_cal)/d(pos_uncal) \approx 0
            #  (which is true for Ch2E \approx linear)
            sigma2_err_cal = abs(
                self.cal.dEdCh(pos_uncal - sigma2_uncal) * sigma2_err_uncal
            )
            return ufloat(sigma2_cal, sigma2_err_cal)
        elif name in ["amp_cal", "eta_cal", "gamma_cal", "vol_cal"]:
            name = name[0 : name.rfind("_cal")]
            return getattr(self, name)
        else:
            # DON'T FORGET THIS LINE! see
            # http://code.activestate.com/recipes/52238/
            raise AttributeError(name)

    def __str__(self):
        return self.formatted_str(verbose=False)

    def formatted_str(self, verbose=False):
        """
        formatted printing of all attributes
        """
        text = ""
        if verbose:
            text += (
                "Pos:         {0.pos_cal:S}\n"
                "Channel:     {0.pos:S}\n"
                "Amp:         {0.amp:S}\n"
                "Sigma1:      {0.sigma1:S}\n"
                "Sigma2:      {0.sigma2:S}\n"
                "Eta:         {0.eta:S}\n"
                "Gamma:       {0.gamma:S}\n"
                "Volume:      {0.vol:S}"
            ).format(self)
        else:
            text += f"Peak@ {self.pos_cal:S}"
        return text

    def __eq__(self, other):
        return self.pos.nominal_value == other.pos.nominal_value

    def __ne__(self, other):
        return self.pos.nominal_value != other.pos.nominal_value

    def __gt__(self, other):
        return self.pos.nominal_value > other.pos.nominal_value

    def __lt__(self, other):
        return self.pos.nominal_value < other.pos.nominal_value

    def __ge__(self, other):
        return self.pos.nominal_value >= other.pos.nominal_value

    def __le__(self, other):
        return self.pos.nominal_value <= other.pos.nominal_value

    def Draw(self, viewport):
        """
        Draw the function of this peak
        """
        if self.viewport:
            if self.viewport == viewport:
                # this has already been drawn
                self.Show()
                return
            else:
                # Unlike the Display object of the underlying implementation,
                # python objects can only be drawn on a single viewport
                raise RuntimeError("Peak cannot be drawn on multiple viewports")
        self.viewport = viewport
        if self.displayObj:
            self.displayObj.Draw(self.viewport)


class PeakModelEE(PeakModel):
    """
    Peak model for electron-electron scattering
    """

    def __init__(self):
        super().__init__()
        self.fParStatus = {
            "pos": None,
            "amp": None,
            "sigma1": None,
            "sigma2": None,
            "eta": None,
            "gamma": None,
            "vol": None,
        }
        # Note that volume is not a true fit parameter, but calculated from
        # the other parameters after the fit
        self.fValidParStatus = {
            "pos": [float, "free", "hold"],
            "amp": [float, "free", "hold"],
            "sigma1": [float, "free", "equal"],
            "sigma2": [float, "free", "equal"],
            "eta": [float, "free", "equal"],
            "gamma": [float, "free", "equal"],
            "vol": ["calculated"],
        }
        self.fOptStatus = {
            "integrate": False,
            "likelihood": "normal",
        }
        self.fValidOptStatus = {
            "integrate": [False, True],
            "likelihood": ["normal", "poisson"],
        }

        self.ResetParamStatus()
        self.name = "ee"
        self.Peak = EEPeak

    def CopyPeak(self, cpeak, color=None, cal=None):
        """
        Copies peak data from a C++ peak class to a Python class
        """
        pos = ufloat(cpeak.GetPos(), cpeak.GetPosError())
        amp = ufloat(cpeak.GetAmp(), cpeak.GetAmpError())
        sigma1 = ufloat(cpeak.GetSigma1(), cpeak.GetSigma1Error())
        sigma2 = ufloat(cpeak.GetSigma2(), cpeak.GetSigma2Error())
        eta = ufloat(cpeak.GetEta(), cpeak.GetEtaError())
        # Triage a bug where error of eta is exceptionally small
        if eta.std_dev / eta.nominal_value < 1e-10:
            eta.std_dev = 0
        gamma = ufloat(cpeak.GetGamma(), cpeak.GetGammaError())
        vol = ufloat(cpeak.GetVol(), cpeak.GetVolError())
        # create the peak object
        peak = self.Peak(pos, amp, sigma1, sigma2, eta, gamma, vol)
        func = cpeak.GetPeakFunc()
        peak.displayObj = ROOT.HDTV.Display.DisplayFunc(func, color)
        peak.displayObj.SetCal(cal)
        return peak

    def RestoreParams(self, peak, cpeak):
        """
        Restore the params of a C++ peak object using a python peak object
        """
        cpeak.RestorePos(peak.pos.nominal_value, peak.pos.std_dev)
        cpeak.RestoreAmp(peak.amp.nominal_value, peak.amp.std_dev)
        cpeak.RestoreSigma1(peak.sigma1.nominal_value, peak.sigma1.std_dev)
        cpeak.RestoreSigma2(peak.sigma2.nominal_value, peak.sigma2.std_dev)
        cpeak.RestoreEta(peak.eta.nominal_value, peak.eta.std_dev)
        cpeak.RestoreGamma(peak.gamma.nominal_value, peak.gamma.std_dev)
        cpeak.RestoreVol(peak.vol.nominal_value, peak.vol.std_dev)

    def ResetParamStatus(self):
        """
        Reset parameter status to defaults
        """
        self.fParStatus["pos"] = "free"
        self.fParStatus["amp"] = "free"
        self.fParStatus["sigma1"] = "equal"
        self.fParStatus["sigma2"] = "equal"
        self.fParStatus["eta"] = "equal"
        self.fParStatus["gamma"] = "equal"
        self.fParStatus["vol"] = "calculated"
        self.fOptStatus["integrate"] = False
        self.fOptStatus["likelihood"] = "normal"

    def Uncal(self, parname, value, pos_uncal, cal):
        """
        Convert a value from calibrated to uncalibrated units
        This is needed, when a value is hold to a specific calibrated value.
        """
        if parname == "pos":
            return cal.E2Ch(value)
        elif parname == "sigma1":
            pos_cal = cal.Ch2E(pos_uncal)
            left_hwhm_uncal = pos_uncal - cal.E2Ch(pos_cal - value)
            return left_hwhm_uncal
        elif parname == "sigma2":
            pos_cal = cal.Ch2E(pos_uncal)
            right_hwhm_uncal = cal.E2Ch(pos_cal + value) - pos_uncal
            return right_hwhm_uncal
        elif parname in ("amp", "eta", "gamma"):
            return value
        else:
            raise RuntimeError("Unexpected parameter name")

    def GetFitter(self, region, peaklist, cal):
        """
        Creates a C++ Fitter object, which can then do the real work
        """

        integrate = self.GetOption("integrate")
        likelihood = self.GetOption("likelihood")
        self.fFitter = ROOT.HDTV.Fit.EEFitter(
            region[0], region[1], integrate, likelihood
        )

        self.ResetGlobalParams()

        # Check if enough values are provided in case of per-peak parameters
        #  (the function raises a RuntimeError if the check fails)
        self.CheckParStatusLen(len(peaklist))

        for pid in range(0, len(peaklist)):
            pos_uncal = peaklist[pid]

            pos = self.GetParam("pos", pid, pos_uncal, cal, pos_uncal)
            amp = self.GetParam("amp", pid, pos_uncal, cal)
            sigma1 = self.GetParam("sigma1", pid, pos_uncal, cal)
            sigma2 = self.GetParam("sigma2", pid, pos_uncal, cal)
            eta = self.GetParam("eta", pid, pos_uncal, cal)
            gamma = self.GetParam("gamma", pid, pos_uncal, cal)

            peak = ROOT.HDTV.Fit.EEPeak(pos, amp, sigma1, sigma2, eta, gamma)
            self.fFitter.AddPeak(peak)

        return self.fFitter



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
HDTV dynamic (C++) library manager
"""

import os
import sys
import shutil
import subprocess
import tempfile
import ROOT
import hdtv.ui
from hdtv.rootext import modules, libfmt

from hdtv._version import get_versions

__version__ = get_versions()["version"]
del get_versions

cachedir = os.path.join(
    os.getenv("XDG_CACHE_HOME", os.path.join(os.environ["HOME"], ".cache")), "hdtv"
)
usrdir = os.path.join(
    cachedir, "%d-%d-%s" % (sys.hexversion, ROOT.gROOT.GetVersionInt(), __version__)
)
sysdir = os.path.join(
    os.path.dirname(__file__), "root-" + str(ROOT.gROOT.GetVersionInt())
)

# Fix performance degradation in ROOT 6.19.02 and higher, caused by
# https://github.com/root-project/root/commit/75161283352217c156c435ba8bac870cd5a125fb
ROOT.gEnv.IgnoreDuplicates(True)
ROOT.gEnv.SetValue("X11.UseXft", 0)


def FindLibrary(name, libname):
    """
    Find the path to a dynamic library in a subfolder.
    Returns the full filename.
    """
    paths = [os.path.join(usrdir, "lib"), os.path.join(sysdir, "lib")]
    for path in paths:
        fname = os.path.join(path, libname)
        if os.path.isfile(fname):
            return fname
    return None


def LoadLibrary(name):
    """
    Load a dynamic library. Try to find and load it, or rebuild it on fail
    """
    loaded = False
    libname = libfmt % name
    fname = FindLibrary(name, libname)
    if fname:
        loaded = _LoadLibrary(fname) >= 0

    if not loaded:
        fname = BuildLibrary(name, usrdir)
        loaded = _LoadLibrary(fname) >= 0

    if not loaded:
        hdtv.ui.error("Failed to load library %s" % libname)
        sys.exit(1)


def _LoadLibrary(fname):
    ROOT.gSystem.SetDynamicPath(
        os.path.dirname(fname) + os.pathsep + ROOT.gSystem.GetDynamicPath()
    )
    ROOT.gSystem.SetIncludePath(
        os.path.dirname(fname) + os.pathsep + ROOT.gSystem.GetDynamicPath()
    )
    return ROOT.gSystem.Load(fname)


def RebuildLibraries(dir, libraries=None):
    if os.path.exists(dir):
        shutil.rmtree(dir)
    for name in libraries or modules:
        BuildLibrary(name, dir)


def BuildLibrary(name, dir):
    srcdir = os.path.join(os.path.dirname(__file__), name)
    # with tempfile.TemporaryDirectory() as tmpdir: TOOD: Not available in python2
    tmpdir = tempfile.mkdtemp()
    subprocess.check_call(
        [
            "cmake",
            srcdir,
            "-DCMAKE_INSTALL_PREFIX=%s" % dir,
            "-DCMAKE_BUILD_TYPE=Release",
        ],
        cwd=tmpdir,
    )
    subprocess.check_call(["cmake", "--build", ".", "-j"], cwd=tmpdir)
    subprocess.check_call(["cmake", "--build", ".", "--target", "install"], cwd=tmpdir)
    shutil.rmtree(tmpdir)

    hdtv.ui.info(f"Rebuild library {(libfmt % name)} in {dir}")
    return os.path.join(dir, "lib", libfmt % name)



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import os

import pytest
from tests.helpers.utils import redirect_stdout, hdtvcmd, isclose, setup_io

from hdtv.util import monkey_patch_ui

monkey_patch_ui()

import hdtv.session

import __main__

try:
    __main__.spectra = hdtv.session.Session()
except RuntimeError:
    pass
session = __main__.spectra

import hdtv.plugins.specInterface
from hdtv.plugins.fitInterface import fit_interface
import hdtv.plugins.rootInterface

import hdtv.cmdline
import hdtv.options
import hdtv.rfile_utils


@pytest.fixture(autouse=True)
def prepare(request):
    fit_interface.ResetFitterParameters()
    # original_wd = os.path.abspath(os.path.join(__file__, os.pardir))
    original_wd = os.getcwd()
    os.chdir(original_wd)
    yield
    os.chdir(original_wd)
    session.Clear()


def test_cmd_root_pwd():
    f, ferr = hdtvcmd("root pwd")
    assert f == os.getcwd()


@pytest.mark.parametrize(
    "start, cd, target",
    [
        ("/", "/tmp", "/tmp"),
        ("/tmp", "..", "/"),
        ("/", "tmp", "/tmp"),
        ("/tmp", "", "/tmp"),
    ],
)
def test_cmd_root_cd(start, cd, target):
    os.chdir(start)
    hdtvcmd("root cd " + cd)
    assert os.getcwd() == target


@pytest.mark.skip(reason="opens TBrowser GUI")
def test_cmd_root_browse():
    hdtvcmd("root browse")
    assert hdtv.plugins.rootInterface.r.browser.GetName() == "Browser"
    hdtv.plugins.rootInterface.r.browser.SetName("Test")
    assert hdtv.plugins.rootInterface.r.browser.GetName() == "Test"
    hdtv.plugins.rootInterface.r.browser = None


# Check that bin sizes are handled in fitted peak volume (see also test_mfile_fit_volume_with_binning)
@pytest.mark.parametrize("spectrum", ["h", "h2"])
@pytest.mark.parametrize(
    "region1, region2, peak, expected_volume", [(500.0, 1500.0, 1000.0, 1000000.0)]
)
def test_root_fit_volume_with_binning(
    region1, region2, peak, expected_volume, spectrum
):
    # Mock parser so RootGet can be used
    # args = mock.Mock()
    args = type("Test", (object,), {})
    args.pattern = [os.path.join("tests", "share", "binning.root", spectrum)]
    args.replace = False
    args.load_cal = False
    args.invisible = False
    hdtv.plugins.rootInterface.r.RootGet(args)

    session.SetMarker("region", region1)
    session.SetMarker("region", region2)
    session.SetMarker("peak", peak)
    session.ExecuteFit()
    fit = session.workFit
    fit_volume = fit.ExtractParams()[0][0]["vol"].nominal_value
    integral_volume = fit.ExtractIntegralParams()[0][0]["vol"].nominal_value

    assert isclose(integral_volume, expected_volume, abs_tol=10)
    assert isclose(fit_volume, expected_volume, abs_tol=200)
    hdtv.plugins.rootInterface.r.RootClose(None)


def test_root_to_root_conversion_for_unconventional_binning():
    # Mock parser so RootGet can be used
    # args = mock.Mock()
    args = type("Test", (object,), {})
    args.replace = False
    args.load_cal = False
    args.invisible = False
    args.pattern = [os.path.join("tests", "share", "binning.root", "h")]
    hdtv.plugins.rootInterface.r.RootGet(args)
    args.pattern = [os.path.join("tests", "share", "binning.root", "h2")]
    hdtv.plugins.rootInterface.r.RootGet(args)

    assert not get_spec(0).cal
    assert isclose(get_spec(1).cal.GetCoeffs()[1], 0.5)


def get_spec(specid):
    s = hdtv.plugins.specInterface.spec_interface
    return s.spectra.dict.get([x for x in list(s.spectra.dict) if x.major == specid][0])



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA


import os

import pytest

from hdtv.util import monkey_patch_ui

monkey_patch_ui()

import __main__

import hdtv.session

try:
    __main__.spectra = hdtv.session.Session()
except RuntimeError:
    pass
spectra = __main__.spectra

from hdtv.plugins.specInterface import spec_interface
from hdtv.plugins.fitInterface import fit_interface
from hdtv.plugins.fitlist import fitxml

testspectrum = os.path.join(os.path.curdir, "tests", "share", "osiris_bg.spc")

test_versions = ["0.1", "1.0", "1.1", "1.3", "1.4", "1.5"]

test_XMLs = [
    os.path.join(os.path.curdir, "tests", "share", "osiris_bg_v" + ver + ".xml")
    for ver in test_versions
]


@pytest.fixture(autouse=True)
def prepare():
    fit_interface.ResetFitterParameters()
    hdtv.options.Set("table", "classic")
    hdtv.options.Set("uncertainties", "short")
    spec_interface.LoadSpectra(testspectrum)
    yield
    spectra.Clear()


@pytest.mark.parametrize("xmlfile", test_XMLs)
def test_old_xml(xmlfile):
    # Check whether hdtv recognizes the format of the XML file at all
    # by invoking the general ReadXML function.
    # Note that this does not ensure the correct reading of the XML file.
    fitxml.ReadXML(spectra.Get("0").ID, xmlfile, refit=True, interactive=False)


"""
print('Saving fits to file %s' % newXML)
fitxml.WriteFitlist(newXML)
print('Deleting all fits')
__main__.spectra.dict["0"].Clear()
print('Reading fits from file %s' % newXML)
fitxml.ReadFitlist(newXML)
fit_interface.ListFits()
"""



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
TextInterface functions
"""

import hdtv.options
import hdtv.ui
import pydoc
from html import escape
import fcntl
import termios
import struct

from prompt_toolkit.application import Application
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.filters import to_filter
from prompt_toolkit.formatted_text import to_formatted_text, fragment_list_to_text, HTML
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout import BufferControl
from prompt_toolkit.layout.containers import HSplit, Window
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.layout.processors import Transformation, Processor


class FormatText(Processor):
    def apply_transformation(self, ti):
        fragments = to_formatted_text(HTML(fragment_list_to_text(ti.fragments)))
        return Transformation(fragments)


class TextInterface(hdtv.ui.SimpleUI):
    def __init__(self, height=25, width=80):
        hdtv.ui.debug("Loaded TextInterface")

        super().__init__()
        # Set options
        self.opt = dict()
        self.opt["ui.pager.cmd"] = hdtv.options.Option(default="less")  # default pager
        self.opt["ui.pager.args"] = hdtv.options.Option(
            default="-F -X -R -S"
        )  # default pager cmd line options
        self.opt["ui.pager.builtin"] = hdtv.options.Option(
            default=True, parse=hdtv.options.parse_bool
        )  # use built-in pager instead of external program
        self.opt["ui.pager.disable_pager"] = hdtv.options.Option(
            default=False, parse=hdtv.options.parse_bool
        )  # whether to not use pager at all

        for (key, opt) in list(self.opt.items()):
            hdtv.options.RegisterOption(key, opt)

        self._fallback_canvasheight = height
        self._fallback_canvaswidth = width
        self._updateTerminalSize()

    def page(self, html, end="\n"):
        """
        Print text by pages
        """
        # TODO: Convert HTML to ANSI?
        if not hdtv.options.Get("ui.pager.builtin"):
            cmd = hdtv.options.Get("ui.pager.cmd")
            args = hdtv.options.Get("ui.pager.args")

            pydoc.tempfilepager(html, str(cmd) + " " + str(args))
        else:
            self.pager(html, end)

    def pager(self, html, end="\n"):
        """
        Construct pager using prompt_toolkit
        """

        my_buffer = Buffer()
        my_window = Window(
            BufferControl(
                buffer=my_buffer,
                focusable=True,
                preview_search=True,
                input_processors=[FormatText()],
            )
        )

        my_buffer.text = html
        my_buffer.read_only = to_filter(True)

        root_container = HSplit(
            [
                my_window,
            ]
        )

        bindings = KeyBindings()

        @bindings.add("c-c")
        @bindings.add("q")
        def _(event):
            "Quit."
            event.app.exit()

        application = Application(
            layout=Layout(
                root_container,
                focused_element=my_window,
            ),
            key_bindings=bindings,
            enable_page_navigation_bindings=True,
            mouse_support=True,
            full_screen=True,
        )

        application.run()
        super().msg(html=html, end=end)

    def msg(self, text=None, *, html=None, end="\n"):
        """
        Message output
        """
        self._updateTerminalSize()
        if not html:
            html = escape(text)
        lines = len(html.splitlines())

        if lines > self.canvasheight and not hdtv.options.Get("ui.pager.disable_pager"):
            self.page(html, end)
        else:
            super().msg(html=html, end=end)

    def _updateTerminalSize(self):
        try:
            h, w, hp, wp = struct.unpack(
                "HHHH",
                fcntl.ioctl(0, termios.TIOCGWINSZ, struct.pack("HHHH", 0, 0, 0, 0)),
            )
            self.canvasheight = h
            self.canvaswidth = w
        except:
            self.canvasheight = 30
            self.canvaswidth = 1050


# initialization
hdtv.ui.ui = TextInterface()
hdtv.ui.msg = hdtv.ui.ui.msg



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
For each model implemented on the C++ side, we have a corresponding Python
class to handle fitter setup and data transfer to the Python side
"""

import ROOT
import hdtv.rootext.display

# Base class for all background models
class BackgroundModel:
    """
    A background model is a function used to fit the quasi-continuous background
    of a spectrum. The user can choose how to fit its parameters (and whether to
    include them at all). After everything has been configured, the background
    model produces a C++ fitter object.
    """

    def __init__(self):
        self.fGlobalParams = dict()
        self.requiredBgRegions = 1

    def ResetGlobalParams(self):
        self.fGlobalParams.clear()

    def OptionsStr(self):
        """
        Returns a string describing the currently set parameters of the model
        """
        statstr = ""

        for name, status in self.fParStatus.items():
            # Short format for multiple values...
            if isinstance(status, list):
                statstr += "%s: " % name
                sep = ""
                for stat in status:
                    statstr += sep
                    if stat in ("free", "equal", "hold", "none", "calculated"):
                        statstr += stat
                    else:
                        statstr += "%.3f" % stat
                    sep = ", "
                statstr += "\n"

            # ... long format for a single value
            else:
                if status == "free":
                    statstr += "%s: (individually) free\n" % name
                elif status == "equal":
                    statstr += "%s: free and equal\n" % name
                elif status == "hold":
                    statstr += "%s: held at default value\n" % name
                elif status == "none":
                    statstr += "%s: none (disabled)\n" % name
                elif status == "calculated":
                    statstr += "%s: calculated from fit result\n" % name
                else:
                    statstr += f"{name}: fixed at {status:.3f}\n"

        return statstr

    def ParseParamStatus(self, parname, status):
        """
        Parse a parameter status specification string
        Raises ValueError when the status is not legal for this parameter
        """
        # Case-insensitive matching
        status = status.strip().lower()

        # Check to see if status corresponds, possibly abbreviated,
        #  to a number of special keywords
        stat = None
        if len(status) < 1:
            raise ValueError
        elif "free"[0 : len(status)] == status:
            stat = "free"

        # If status was a keyword, see if this setting is legal for the
        # parameter
        if stat is not None:
            if stat not in self.fValidParStatus[parname]:
                msg = "Status {} not allowed for parameter {} in peak model {}".format(
                    stat,
                    parname,
                    self.name,
                )
                raise ValueError(msg)
            return stat

        # If status was not a keyword, try to parse it as an integer. If that
        # fails, we are out of options.
        try:
            val = int(status)
        except ValueError:
            msg = "Failed to parse status specifier `%s'" % status
            raise ValueError(msg)

        # Check if a numeric value is legal for the parameter
        if int not in self.fValidParStatus[parname]:
            msg = "Invalid status {} for parameter {} in peak model {}".format(
                status,
                parname,
                self.name,
            )
            raise ValueError(msg)
        return val

    def SetParameter(self, parname, status):
        """
        Set status for a certain parameter. Status is a string describing the
        desired status. Raises ValueError in case of invalid input.

        status may be a string or a number, depending on the parameter type
        """
        parname = parname.strip().lower()

        if parname not in list(self.fValidParStatus.keys()):
            raise ValueError(
                "Invalid parameter name %s for background model %s"
                % (parname, self.name)
            )

        self.fParStatus[parname] = self.ParseParamStatus(parname, status)



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2010  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import argparse
from uncertainties import ufloat_fromstr

import hdtv.cmdline
import hdtv.ui
import hdtv.util


class FitMap:
    def __init__(self, spectra, ecal):
        hdtv.ui.debug("Loaded plugin for setting nominal positions to peaks")

        self.spectra = spectra
        self.ecal = ecal

        prog = "fit position assign"
        description = "assign energy value as nominal position for peak"
        usage = "%(prog)s peakid energy [peakid energy ...] "
        parser = hdtv.cmdline.HDTVOptionParser(
            prog=prog, description=description, usage=usage
        )
        parser.add_argument(
            "args",
            metavar="peakid energy",
            help="peak energy pairs",
            nargs=argparse.REMAINDER,
        )
        hdtv.cmdline.AddCommand(prog, self.FitPosAssign, parser=parser)

        prog = "fit position erase"
        description = "erase nominal position for peaks"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument("peakid", nargs="+", help="peak id")
        hdtv.cmdline.AddCommand(prog, self.FitPosErase, parser=parser)

        prog = "fit position map"
        description = "read nominal position from file"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument("filename")
        parser.add_argument(
            "-o",
            "--overwrite",
            action="store_true",
            default=False,
            help="Overwrite existing nominal peak positions",
        )
        parser.add_argument(
            "-t",
            "--tolerance",
            default=12,
            type=float,
            help="Tolerance for associating peaks to positions [default:%(default)s]",
        )
        hdtv.cmdline.AddCommand(prog, self.FitPosMap, fileargs=True, parser=parser)

        prog = "calibration position recalibrate"
        description = (
            "use stored nominal positions of peaks to (re)calibrate the spectrum"
        )
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-s",
            "--spectrum",
            action="store",
            default="active",
            help="spectrum ids to apply calibration to",
        )
        parser.add_argument(
            "-d",
            "--degree",
            action="store",
            default="1",
            help="degree of calibration polynomial fitted [default: %(default)s]",
        )
        parser.add_argument(
            "-f",
            "--show-fit",
            action="store_true",
            default=False,
            help="show fit used to obtain calibration",
        )
        parser.add_argument(
            "-r",
            "--show-residual",
            action="store_true",
            default=False,
            help="show residual of calibration fit",
        )
        parser.add_argument(
            "-t",
            "--show-table",
            action="store_true",
            default=False,
            help="print table of energies given and energies obtained from fit",
        )
        parser.add_argument(
            "-i",
            "--ignore-errors",
            action="store_true",
            default=False,
            help="set all weights to 1 in fit (ignore error bars even if given)",
        )
        parser.add_argument(
            "fitids",
            nargs="*",
            default=["all"],
            help="ids of the fits/peaks to use for the calibration (default=all)",
        )
        hdtv.cmdline.AddCommand(prog, self.CalPosRecalibrate, parser=parser)

    def FitPosAssign(self, args):
        """
        Assign a nominal value for the positon of peaks

        Peaks are specified by their id and the peak number within the fit.
        Syntax: id.number
        If no number is given, the first peak in the fit is used.
        """
        if self.spectra.activeID is None:
            hdtv.ui.warning("No active spectrum, no action taken.")
            return False
        spec = self.spectra.GetActiveObject()
        if len(args.args) % 2 != 0:
            raise hdtv.cmdline.HDTVCommandError("Number of arguments must be even")
        else:
            for i in range(0, len(args.args), 2):
                en = ufloat_fromstr(args.args[i + 1])
                try:
                    ids = hdtv.util.ID.ParseIds(args.args[i], spec, only_existent=False)
                    pid = ids[0].minor
                    if pid is None:
                        pid = 0
                    fid = ids[0]
                    fid.minor = None
                    spec.dict[fid].peaks[pid].extras["pos_lit"] = en
                except ValueError:
                    continue
                except (KeyError, IndexError):
                    hdtv.ui.warning("no peak with id %s" % args.args[i])

    def FitPosErase(self, args):
        """
        Erase nominal values for the position of peaks

        Peaks are specified by their id and the peak number within the fit.
        Syntax: id.number
        If no number is given, the all peaks in that fit are used.
        """
        if self.spectra.activeID is None:
            hdtv.ui.warning("No active spectrum, no action taken.")
            return False
        spec = self.spectra.GetActiveObject()
        # use set, to prevent duplicates
        pids = set()
        fids = hdtv.util.ID.ParseIds(args.peakid, spec, only_existent=False)
        for i in fids:
            if i.minor is not None:
                pids.add(i)
            else:
                try:
                    # collect all peaks in this fit
                    pids.update(set(hdtv.util.ID.ParseIds("all", spec.dict[i])))
                except KeyError:
                    hdtv.ui.warning("no peak with id %s" % i.major)
                    continue
        for i in pids:
            pid = i.minor
            i.minor = None
            try:
                peak = spec.dict[i].peaks[pid]
            except (KeyError, IndexError):
                # ignore invalid peaks, but give a warning
                hdtv.ui.warning(f"no peak with id {i.major}.{pid}")
                continue
            try:
                peak.extras.pop("pos_lit")
            except KeyError:
                # ignore peaks where "pos_lit" is unset
                continue

    def FitPosMap(self, args):
        """
        Read a list of energies from file and map to the fitted peaks.

        The spectrum must be roughly calibrated for this to work.
        """
        f = hdtv.util.TxtFile(args.filename)
        f.read()
        energies = list()
        for line in f.lines:
            energies.append(ufloat_fromstr(line.split(",")[0]))
        if self.spectra.activeID is None:
            hdtv.ui.warning("No active spectrum, no action taken.")
            return False
        if len(energies) == 0:
            hdtv.ui.warning(f"No energies found in file {args.filename}.")
            return False
        spec = self.spectra.GetActiveObject()
        count = 0
        for fit in spec.dict.values():
            for peak in fit.peaks:
                if args.overwrite:
                    peak.extras.pop("pos_lit", None)
                # pick best match within a certain tolerance (if it exists)
                tol = args.tolerance
                enlit = [e for e in energies if abs(peak.pos_cal.std_score(e)) < tol]
                if len(enlit) > 0:
                    peak.extras["pos_lit"] = min(
                        enlit, key=lambda e: abs(peak.pos_cal.std_score(e))
                    )
                    count += 1
        # give a feetback to the user
        hdtv.ui.msg("Mapped %s energies to peaks" % count)

    def CalPosRecalibrate(self, args):
        if self.spectra.activeID is None:
            hdtv.ui.warning("No active spectrum, no action taken.")
            return False
        spec = self.spectra.GetActiveObject()
        # parsing of command line
        sids = hdtv.util.ID.ParseIds(args.spectrum, self.spectra)
        if len(sids) == 0:
            sids = [self.spectra.activeID]
        degree = int(args.degree)
        fitIDs = hdtv.util.ID.ParseIds(args.fitids, spec, only_existent=False)
        peakDict = {}
        for fitID in fitIDs:
            try:
                if fitID.minor is None:
                    for minor, peak in enumerate(spec.dict[fitID].peaks):
                        peakDict[hdtv.util.ID(fitID.major, minor)] = peak
                else:
                    peakDict[fitID] = spec.dict[hdtv.util.ID(fitID.major)].peaks[
                        fitID.minor
                    ]
            except (IndexError, KeyError):
                hdtv.ui.warning(f"Ignoring invalid fit/peak id {fitID}")
        pairs = hdtv.util.Pairs()
        for peakId, peak in peakDict.items():
            if "pos_lit" in peak.extras:
                enlit = peak.extras["pos_lit"]
                pairs.add(peak.pos, enlit)
            elif args.fitids != ["all"]:
                hdtv.ui.warning(
                    f"Ignoring fit/peak id {peakId} without assigned nominal position"
                )
        cal = self.ecal.CalFromPairs(
            pairs,
            degree,
            table=args.show_table,
            fit=args.show_fit,
            residual=args.show_residual,
            ignore_errors=args.ignore_errors,
        )
        self.spectra.ApplyCalibration(sids, cal)
        return True


# plugin initialisation
import __main__
from hdtv.plugins.calInterface import energy_cal_interface

hdtv.cmdline.RegisterInteractive(
    "fitmap", FitMap(__main__.spectra, energy_cal_interface)
)



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

from array import array
from html import escape
import re
import fnmatch

import ROOT
import hdtv.util
import hdtv.cal
import hdtv.options

import hdtv.rootext.calibration


class PositionCalibrationDict(dict):
    """
    Dict subclass that stores position calibrations

    Subclasses dict overwriting get() and providing __missing__() and a user config option to enable calibration lookups based on glob or regex matching of dict keys to lookup key.
    Note that in all modes a literally matching dict key is always preferred to globs/regex.
    Note that other methods like __contains__() are not overwritten on purpose, so use "key in dict" to check for literally existing keys and dict[key] or get() to check for glob/regex matching keys.
    """

    calListSpecNameMatching = hdtv.options.Option(
        default="glob", parse=hdtv.options.parse_choices(["literal", "glob", "regex"])
    )
    hdtv.options.RegisterOption(
        "calibration.position.list.spec_name_matching", calListSpecNameMatching
    )

    def __missing__(self, key):
        matchMode = self.calListSpecNameMatching.Get()
        if matchMode == "glob":
            for dictKey, val in self.items():
                if fnmatch.fnmatch(key, dictKey):
                    return val
        elif matchMode == "regex":
            for dictKey, val in self.items():
                if re.fullmatch(dictKey, key):
                    return val
        raise KeyError(key)

    def get(self, key, default=None):
        """
        Return the value for key if key is in the dictionary, else the value of a glob/regex (if enabled) matching key, else default.
        """
        try:
            return self[key]
        except KeyError:
            return default


def MakeCalibration(cal):
    """
    Create a ROOT.HDTV.Calibration object from a python list
    """
    if not isinstance(cal, ROOT.HDTV.Calibration):
        if cal is None:
            cal = []  # Trivial calibration, degree -1
        calarray = ROOT.TArrayD(len(cal))
        for i, c in enumerate(cal):
            calarray[i] = c
        # create the calibration object
        cal = ROOT.HDTV.Calibration(calarray)
    return cal


def GetCoeffs(cal):
    """
    Get the list of calibration coeffs from the ROOT.HDTV.Calibration object
    """
    return list(cal.GetCoeffs())


def PrintCal(cal):
    """
    Get the calibration as string
    """
    return "   ".join([str(c) for c in GetCoeffs(cal)])


class CalibrationFitter:
    """
    Fit a calibration polynom to a list of channel/energy pairs.
    """

    def __init__(self):
        self.Reset()

    def Reset(self):
        self.pairs = []
        self.calib = None
        self.chi2 = None
        self.__TF1 = None
        self.__TF1_id = None
        self.graph = ROOT.TGraph()

    def AddPair(self, ch, e):
        self.pairs.append([ch, e])

    def FitCal(self, degree, ignore_errors=False):
        """
        Use the reference peaks found in the histogram to fit the actual
        calibration function.
        If degree == 0, the linear coefficient of the polynomial is fixed at 1.

        If self.pairs contains ufloats the channel std_dev is respected
        """
        hdtv.ui.debug("conducting calibration from fitted peaks")
        if degree < 0:
            raise ValueError("Degree cannot be negative")

        if len(self.pairs) < degree + 1:
            raise RuntimeError(
                "You must specify at least as many channel/energy pairs as there are free parameters"
            )

        self.__TF1_id = "calfitter_" + hex(id(self))  # unique function ID

        # Create ROOT function
        if degree == 0:
            self.__TF1 = ROOT.TF1(self.__TF1_id, "pol1", 0, 0)
            self.__TF1.FixParameter(1, 1.0)
            degree = 1
        else:
            self.__TF1 = ROOT.TF1(self.__TF1_id, "pol%d" % degree, 0, 0)

        # Prepare data for fitter
        channels = array("d")
        channels_err = array("d")
        energies = array("d")
        energies_err = array("d")
        all_have_error = True
        any_has_error = False
        any_has_xerror = False

        for (ch, e) in self.pairs:
            has_xerror = False
            has_error = False

            # Store channels
            try:  # try to read from ufloat
                channel = float(ch.nominal_value)
                channel_err = float(ch.std_dev)
                channels.append(channel)
                channels_err.append(channel_err)
                if ch.std_dev != 0.0:
                    has_xerror = True
            except AttributeError:
                channels.append(float(ch))
                channels_err.append(0.0)

            any_has_xerror = has_xerror or any_has_xerror

            # Store energies
            try:  # try to read from ufloat
                energy = float(e.nominal_value)
                energy_err = float(e.std_dev)
                energies.append(energy)
                energies_err.append(energy_err)
                if e.std_dev != 0.0:
                    has_error = True
            except AttributeError:
                energies.append(float(e))
                energies_err.append(0.0)

            all_have_error = all_have_error and has_error
            any_has_error = any_has_error or has_error

        hdtv.ui.debug("all_have_error: " + str(all_have_error), level=2)
        hdtv.ui.debug("any_has_error: " + str(any_has_error), level=2)
        hdtv.ui.debug("any_has_xerror: " + str(any_has_xerror), level=2)

        hdtv.ui.debug("channels: " + str(channels), level=5)
        hdtv.ui.debug("channels err: " + str(channels_err), level=5)
        hdtv.ui.debug("energies: " + str(energies), level=5)
        hdtv.ui.debug("energies err: " + str(energies_err), level=5)

        self.__TF1.SetRange(0, max(energies) * 1.1)
        self.TGraph = ROOT.TGraphErrors(
            len(energies), channels, energies, channels_err, energies_err
        )

        fitoptions = "0"  # Do not plot
        fitoptions += "Q"  # Quiet
        fitoptions += "S"  # Return TFitResult for new ROOT versions

        if not ignore_errors and not all_have_error:
            ignore_errors = True
            if any_has_error:
                hdtv.ui.warning(
                    "Some values specified without error, ignoring all errors in fit"
                )

        if ignore_errors:
            fitoptions += "W"
        else:
            hdtv.ui.info("doing error-weighted fit")
            if any_has_xerror:
                # We must use the iterative fitter (minuit) to take x errors
                # into account.
                fitoptions += "F"
                hdtv.ui.info(
                    "switching to non-linear fitter (minuit) for x error weighting"
                )

        # Do the fit
        result = self.TGraph.Fit(self.__TF1_id, fitoptions)
        if isinstance(result, int):
            if result != 0:
                raise RuntimeError("Fit failed")
        elif isinstance(result, ROOT.TFitResultPtr):
            if result.Get().Status() != 0:
                raise RuntimeError("Fit failed")
        else:  # Fallback, attempt to cast result to an int
            if int(result) != 0:
                raise RuntimeError("Fit failed")

        # Save the fit result
        self.calib = MakeCalibration(
            [self.__TF1.GetParameter(i) for i in range(0, degree + 1)]
        )
        self.chi2 = self.__TF1.GetChisquare()

    def ResultStr(self):
        """
        Return string describing the result of the calibration
        """
        if self.calib is None:
            raise RuntimeError("No calibration available (did you call FitCal()?)")

        s = "<b>Calibration</b>: "
        s += " ".join([escape("%.6e") % x for x in self.calib.GetCoeffs()])
        s += "\n<b>Chi²</b>: %.4f" % self.chi2

        return s

    def ResultTable(self):
        """
        Return a table showing the fit results
        """
        if self.calib is None:
            raise RuntimeError("No calibration available (did you call FitCal()?)")

        header = ["Channel", "E_given", "E_fit", "Residual"]
        keys = "channel", "e_given", "e_fit", "residual"
        tabledata = list()

        for (ch, e_given) in self.pairs:

            tableline = dict()
            e_fit = self.calib.Ch2E(ch.nominal_value)
            residual = e_given - e_fit

            tableline["channel"] = "%10.2f" % ch.nominal_value
            tableline["e_given"] = "%10.2f" % e_given.nominal_value
            tableline["e_fit"] = "%10.2f" % e_fit
            tableline["residual"] = "%10.2f" % residual.nominal_value
            tabledata.append(tableline)

        return hdtv.util.Table(tabledata, keys, header=header, sortBy="channel")

    def DrawCalFit(self):
        """
        Draw fit used for calibration
        """
        if self.calib is None:
            raise RuntimeError("No calibration available (did you call FitCal()?)")

        canvas = ROOT.TCanvas("CalFit", "Calibration Fit")
        # Prevent canvas from being closed as soon as this function finishes
        ROOT.SetOwnership(canvas, False)

        min_ch = self.pairs[0][0].nominal_value
        max_ch = self.pairs[0][0].nominal_value
        graph = ROOT.TGraphErrors(len(self.pairs))
        ROOT.SetOwnership(graph, False)

        i = 0
        for (ch, e) in self.pairs:
            min_ch = min(min_ch, ch.nominal_value)
            max_ch = max(max_ch, ch.nominal_value)

            try:
                graph.SetPoint(i, ch.nominal_value, e.nominal_value)
            except BaseException:
                graph.SetPoint(i, ch.nominal_value, e)
            graph.SetPointError(i, ch.std_dev, 0)
            i += 1

        coeffs = self.calib.GetCoeffs()
        func = ROOT.TF1("CalFitFunc", "pol%d" % (len(coeffs) - 1), min_ch, max_ch)
        ROOT.SetOwnership(func, False)
        for i, coeff in enumerate(coeffs):
            func.SetParameter(i, coeff)

        graph.SetTitle("Calibration Fit")
        graph.GetXaxis().SetTitle("Channel")
        graph.GetXaxis().CenterTitle()
        graph.GetYaxis().SetTitle("Energy")
        graph.GetYaxis().CenterTitle()

        graph.Draw("A*")
        func.Draw("SAME")
        canvas.Update()

    def DrawCalResidual(self):
        """
        Debug: draw residual of fit used for calibration
        """
        if self.calib is None:
            raise RuntimeError("No calibration available (did you call FitCal()?)")

        canvas = ROOT.TCanvas("CalResidual", "Calibration Residual")
        # Prevent canvas from being closed as soon as this function finishes
        ROOT.SetOwnership(canvas, False)

        min_ch = self.pairs[0][0].nominal_value
        max_ch = self.pairs[0][0].nominal_value
        graph = ROOT.TGraph(len(self.pairs))
        ROOT.SetOwnership(graph, False)

        i = 0
        for (ch, e) in self.pairs:
            min_ch = min(min_ch, ch.nominal_value)
            max_ch = max(max_ch, ch.nominal_value)
            try:
                # energie may be ufloat
                e = e.nominal_value
            except BaseException:
                pass

            graph.SetPoint(i, ch.nominal_value, e - self.calib.Ch2E(ch.nominal_value))
            i += 1

        nullfunc = ROOT.TF1("CalResidualFunc", "pol0", min_ch, max_ch)
        ROOT.SetOwnership(nullfunc, False)

        graph.SetTitle("Residual of calibration fit")
        graph.GetXaxis().SetTitle("Channel")
        graph.GetXaxis().CenterTitle()
        graph.GetYaxis().SetTitle("Energy difference")
        graph.GetYaxis().CenterTitle()

        graph.Draw("A*")
        nullfunc.Draw("SAME")
        canvas.Update()



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2019  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
This is a dummy/noop module that can replace ROOT.HDTV.Display
without actually doing anything. In particular, no GUI windows
are opened.
"""


def _noop(*args, **kwargs):
    pass


def _get_float(self, *args, **kwargs):
    self.i += 0.1
    return self.i


def _get_bool(self, *args, **kwargs):
    return True


class Viewer:
    Connect = _noop

    def __init__(self, *args, **kwargs):
        self.fView = View1D()
        self.fKeySym = 0
        self.fKeyStr = ""

    def GetViewport(self):
        return self.fView


class View1D:
    GetCursorX = _get_float
    GetCursorY = _get_float
    GetDarkMode = _get_bool
    GetXOffset = _get_float
    GetXVisibleRegion = _get_float
    GetYMinVisibleRegion = _get_float
    GetYOffset = _get_float
    GetYVisibleRegion = _get_float
    LockUpdate = _noop
    SetDarkMode = _noop
    SetStatusText = _noop
    SetXCenter = _noop
    SetXVisibleRegion = _noop
    SetYMinVisibleRegion = _noop
    SetYVisibleRegion = _noop
    ShiftXOffset = _noop
    ShiftYOffset = _noop
    ShowAll = _noop
    ToggleLogScale = _noop
    ToggleUseNorm = _noop
    ToggleYAutoScale = _noop
    UnlockUpdate = _noop
    Update = _noop
    XZoomAroundCursor = _noop
    YAutoScaleOnce = _noop
    YZoomAroundCursor = _noop

    def __init__(self, *args, **kwargs):
        self.i = 0.1


class DisplayObj:
    Draw = _noop
    Hide = _noop
    Show = _noop
    __call__ = _noop

    def __init__(self, *args, **kwargs):
        self.ID = None


class DisplayBlock(DisplayObj):
    SetCal = _noop
    SetColor = _noop
    SetNorm = _noop


class Marker(DisplayObj):
    SetColor = _noop
    SetDash = _noop
    SetID = _noop
    SetN = _noop
    SetPos = _noop


class XMarker(Marker):
    SetCal = _noop
    SetConnectTop = _noop


class YMarker(Marker):
    pass


class DisplayFunc(DisplayBlock):
    pass


class DisplaySpec(DisplayBlock):
    SetHist = _noop
    SetID = _noop


class MTViewer(Viewer):
    pass



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
Write and Read Fitlist saved in xml format
"""

import os
import glob
import hdtv.cmdline
import hdtv.options
import hdtv.fitxml
import hdtv.ui
import hdtv.util


class FitlistManager:
    def __init__(self, spectra):
        hdtv.ui.debug("Loaded fitlist plugin")

        self.spectra = spectra
        self.xml = hdtv.fitxml.FitXml(spectra)
        self.list = dict()

        self.tv = FitlistHDTVInterface(self)

    def WriteXML(self, sid, fname=None):
        name = self.spectra.dict[sid].name
        # remember absolute pathname for later use
        fname = os.path.abspath(fname)
        self.list[name] = fname
        with hdtv.util.open_compressed(fname, mode="wb") as f:
            self.xml.WriteFitlist(f, sid)

    def ReadXML(
        self, sid, fname, calibrate=False, refit=False, interactive=True, associate=True
    ):
        spec = self.spectra.dict[sid]
        # remember absolute pathname for later use
        fname = os.path.abspath(fname)
        if associate:
            self.list[spec.name] = fname
        else:
            self.list.pop(spec.name, None)
        with hdtv.util.open_compressed(fname, mode="rb") as f:
            self.xml.ReadFitlist(
                f,
                sid,
                calibrate=calibrate,
                refit=refit,
                interactive=interactive,
                fname=fname,
            )

    def WriteList(self, fname):
        lines = list()
        listpath = os.path.abspath(fname)
        for (spec, xml) in self.list.items():
            # create relativ path name
            common = os.path.commonprefix([listpath, xml])
            xml = xml.replace(common, "")
            xml = xml.strip("/")
            lines.append(spec + ": " + xml)
        text = "\n".join(lines)
        with open(fname, "w") as f:
            f.write(text)

    def ReadList(self, fname):
        with open(fname) as f:
            dirname = os.path.dirname(fname)
            for linenum, l in enumerate(f):
                # Remove comments and whitespace; ignore empty lines
                l = l.split("#", 1)[0].strip()
                if l == "":
                    continue
                try:
                    (k, v) = l.split(":", 1)
                    name = k.strip()
                    xmlfile = v.strip()
                    # create valid path from relative pathnames
                    xmlfile = os.path.join(dirname, xmlfile)
                    if not os.path.exists(xmlfile):
                        hdtv.ui.warning("No such file %s" % xmlfile)
                        continue
                    sid = None
                    for ID in self.spectra.ids:
                        if self.spectra.dict[ID].name == name:
                            sid = ID
                            break
                    if sid is not None:
                        self.ReadXML(sid, xmlfile)
                    else:
                        hdtv.ui.warning("Spectrum %s is not loaded. " % name)
                except ValueError:
                    hdtv.ui.warning(
                        "Could not parse line %d of file %s: ignored."
                        % (linenum + 1, fname)
                    )


class FitlistHDTVInterface:
    def __init__(self, FitlistIf):
        self.FitlistIf = FitlistIf
        self.spectra = FitlistIf.spectra

        prog = "fit write"
        description = "write fits to xml file"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-s",
            "--spectrum",
            action="store",
            default="active",
            help="for which the fits should be saved (default=active)",
        )
        parser.add_argument(
            "-F",
            "--force",
            action="store_true",
            default=False,
            help="overwrite existing files without asking",
        )
        parser.add_argument(
            "filename",
            nargs="?",
            default=None,
            help="""may contain %%s, %%d, %%02d (or other python
            format specifier) as placeholder for spectrum id, if not provided defaults to the spectrum's associated fitlist""",
        )
        hdtv.cmdline.AddCommand(prog, self.FitWrite, fileargs=True, parser=parser)

        prog = "fit read"
        description = "read fits from xml file"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-s",
            "--spectrum",
            action="store",
            default="active",
            help="spectra to which the fits should be added (default=active)",
        )
        parser.add_argument(
            "-r",
            "--refit",
            action="store_true",
            default=False,
            help="Force refitting during load",
        )
        parser.add_argument(
            "-c",
            "--calibrate",
            action="store_true",
            default=False,
            help="Apply the stored calibration to the loaded spectrum",
        )
        parser.add_argument(
            "-n",
            "--no-associate",
            action="store_true",
            default=False,
            help="""Do not remeber the filename of the fitlist file,
            i.e. `fit write` will not write to the original fitlist file,
            but create a new one according to the name of the spectrum.
            Useful for reusing fits from a different spectrum.""",
        )
        parser.add_argument(
            "filename",
            nargs="*",
            help="""May contain %%s, %%d, %%02d (or other python
            format specifier) as placeholder for spectrum id, if not provided defaults to each spectrum's associated fitlist""",
        )
        hdtv.cmdline.AddCommand(prog, self.FitRead, fileargs=True, parser=parser)

        prog = "fit getlists"
        description = "reads fitlists according to the list saved in a file"
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument("filename", default=None)
        hdtv.cmdline.AddCommand(prog, self.FitGetlists, fileargs=True, parser=parser)

        prog = "fit savelists"
        description = (
            "saves a list of spectrum names and corresponding fitlist files to file"
        )
        parser = hdtv.cmdline.HDTVOptionParser(prog=prog, description=description)
        parser.add_argument(
            "-F",
            "--force",
            action="store_true",
            default=False,
            help="overwrite existing files without asking",
        )
        parser.add_argument("filename", metavar="output-file", default=None)
        hdtv.cmdline.AddCommand(prog, self.FitSavelists, fileargs=True, parser=parser)

    def FitWrite(self, args):
        """
        Saving a fitlist as xml
        """
        # TODO: this need urgent cleanup and testing, especially for the saving
        # of fitlists from multiple spectra, but I'm really in a hurry now. Sorry. Ralf
        # get spectrum
        sids = hdtv.util.ID.ParseIds(args.spectrum, __main__.spectra)
        if len(sids) == 0:
            raise hdtv.cmdline.HDTVCommandError("There is no active spectrum")
        if len(sids) > 1:
            # TODO: Check if placeholder character is present in filename and
            # warn if not
            pass
        #            raise hdtv.cmdline.HDTVCommandError("Can only save fitlist of one spectrum")
        for sid in sids:
            #            sid = sids[0]
            # get filename
            if args.filename is None:
                name = self.spectra.dict[sid].name
                try:
                    fname = self.FitlistIf.list[name]
                except KeyError:
                    (base, ext) = os.path.splitext(name)
                    fname = base + "." + hdtv.options.Get("fit.list.default_extension")
            else:
                fname = os.path.expanduser(args.filename)
                # Try to replace placeholder "%s" in filename with specid
                try:
                    fname = fname % sid
                except TypeError:  # No placeholder found
                    # TODO: do something sensible here... Luckily hdtv will not
                    # overwrite spectra without asking...
                    pass
            hdtv.ui.msg("Saving fits of spectrum %d to %s" % (sid, fname))

            if hdtv.util.user_save_file(fname, args.force):
                self.FitlistIf.WriteXML(sid, fname)

    def FitRead(self, args):
        """
        reading a fitlist from xml
        """
        fnames = dict()  # Filenames for each spectrum ID
        sids = hdtv.util.ID.ParseIds(args.spectrum, __main__.spectra)
        if len(sids) == 0:
            raise hdtv.cmdline.HDTVCommandError("There is no active spectrum")

        # Build list of files to load for each spectrum
        for sid in sids:
            if not args.filename:
                name = self.spectra.dict[sid].name
                try:
                    fnames[sid] = (self.FitlistIf.list[name],)
                except KeyError:
                    (base, _) = os.path.splitext(name)
                    fname = base + "." + hdtv.options.Get("fit.list.default_extension")
                    if not os.path.isfile(fname):
                        hdtv.ui.warning("No such file %s" % fname)
                    else:
                        fnames[sid] = (fname,)
            else:
                fnames[sid] = list()  # Filenames for this spectrum ID
                for fname_raw in args.filename:
                    try:
                        # Try to replace format placeholder (e.g. %s) with spectrum
                        # ID
                        fname = fname_raw % sid
                    except TypeError:  # No placeholder found
                        fname = fname_raw

                    fname = os.path.expanduser(fname)
                    more = glob.glob(fname)
                    if len(more) == 0:
                        hdtv.ui.warning("No such file %s" % fname)
                    fnames[sid].extend(more)

        # Load files
        for sid, fnameList in fnames.items():
            for fname in fnameList:
                hdtv.ui.msg(f"Reading fitlist {fname} to spectrum {sid}")
                self.FitlistIf.ReadXML(
                    sid,
                    fname,
                    calibrate=args.calibrate,
                    refit=args.refit,
                    associate=(not args.no_associate),
                )

    def FitSavelists(self, args):
        if hdtv.util.user_save_file(args.filename, args.force):
            self.FitlistIf.WriteList(args.filename)

    def FitGetlists(self, args):
        fname = glob.glob(os.path.expanduser(args.filename))
        if len(fname) > 1:
            raise hdtv.cmdline.HDTVCommandError("More than 1 files match the pattern")
        fname = fname[0]
        if not os.path.exists(fname):
            raise hdtv.cmdline.HDTVCommandError("No such file %s" % fname)
        self.FitlistIf.ReadList(fname)


hdtv.options.RegisterOption(
    "fit.list.default_extension", hdtv.options.Option(default="xfl")
)

import __main__

fitxml = FitlistManager(__main__.spectra)
hdtv.cmdline.RegisterInteractive("fitxml", fitxml)



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2018  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

from uncertainties import ufloat

import ROOT
import hdtv.rootext.fit


def Integrate(spec, bg, region):
    hist = spec.hist.hist
    region.sort()

    int_tot = ROOT.HDTV.Fit.TH1Integral(hist, region[0], region[1])
    int_bac = (
        ROOT.HDTV.Fit.BgIntegral(bg, region[0], region[1], hist.GetXaxis())
        if bg
        else None
    )
    int_sub = (
        ROOT.HDTV.Fit.TH1BgsubIntegral(hist, bg, region[0], region[1]) if bg else None
    )

    return {
        kind: get_integral_info(spec, integral)
        for (kind, integral) in zip(("tot", "bg", "sub"), (int_tot, int_bac, int_sub))
    }


def get_integral_info(spec, integral):
    if not integral:
        return None
    pos = ufloat(integral.GetMean(), abs(integral.GetMeanError()))
    width = ufloat(integral.GetWidth(), abs(integral.GetWidthError()))
    vol = ufloat(integral.GetIntegral(), abs(integral.GetIntegralError()))
    skew = ufloat(integral.GetRawSkewness(), abs(integral.GetRawSkewnessError()))
    integral_info = {"uncal": {"pos": pos, "width": width, "vol": vol, "skew": skew}}

    if spec.cal:
        return calibrate_integral(integral_info, spec.cal)
    return integral_info


def calibrate_integral(integral_info, cal):
    pos = integral_info["uncal"]["pos"]
    width = integral_info["uncal"]["width"]

    pos_cal = ufloat(
        cal.Ch2E(pos.nominal_value), abs(cal.dEdCh(pos.nominal_value) * pos.std_dev)
    )
    hwhm_uncal = width.nominal_value / 2
    width_cal_n = cal.Ch2E(pos.nominal_value + hwhm_uncal) - cal.Ch2E(
        pos.nominal_value - hwhm_uncal
    )
    # This is only an approximation, valid as d(width_cal_n)/d(pos.nominal_value) ≈ 0
    #  (which is true for Ch2E ≈ linear)
    width_cal_std_dev = abs(
        (
            cal.dEdCh(pos.nominal_value + hwhm_uncal) / 2.0
            + cal.dEdCh(pos.nominal_value - hwhm_uncal) / 2.0
        )
        * width.std_dev
    )
    width_cal = ufloat(width_cal_n, abs(width_cal_std_dev))
    # TODO: Does it make sense to calibrate the skewness?
    # Not relevant for calibrations with degree < 2

    cal = {"pos": pos_cal, "width": width_cal, "vol": integral_info["uncal"]["vol"]}
    integral_info.update({"cal": cal})

    return integral_info



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import ROOT


class FitPanel:
    def __init__(self):
        self._dispatchers = []
        self.fFitHandler = None
        self.fClearHandler = None
        self.fResetHandler = None
        self.fDecompHandler = None
        self.fVisible = False

        self.fMainFrame = ROOT.TGMainFrame(ROOT.gClient.GetRoot(), 300, 500)
        self.fMainFrame.DontCallClose()
        disp = ROOT.TPyDispatcher(self.Hide)
        self.fMainFrame.Connect("CloseWindow()", "TPyDispatcher", disp, "Dispatch()")
        self._dispatchers.append(disp)

        ## Button frame ##
        self.fButtonFrame = ROOT.TGHorizontalFrame(self.fMainFrame)

        self.fFitButton = ROOT.TGTextButton(self.fButtonFrame, "Fit")
        disp = ROOT.TPyDispatcher(self.FitClicked)
        self.fFitButton.Connect("Clicked()", "TPyDispatcher", disp, "Dispatch()")
        self._dispatchers.append(disp)
        self.fButtonFrame.AddFrame(self.fFitButton)

        self.fClearButton = ROOT.TGTextButton(self.fButtonFrame, "Clear")
        disp = ROOT.TPyDispatcher(self.ClearClicked)
        self.fClearButton.Connect("Clicked()", "TPyDispatcher", disp, "Dispatch()")
        self._dispatchers.append(disp)
        self.fButtonFrame.AddFrame(self.fClearButton)

        self.fResetButton = ROOT.TGTextButton(self.fButtonFrame, "Reset")
        disp = ROOT.TPyDispatcher(self.ResetClicked)
        self.fResetButton.Connect("Clicked()", "TPyDispatcher", disp, "Dispatch()")
        self._dispatchers.append(disp)
        self.fButtonFrame.AddFrame(self.fResetButton)

        self.fHideButton = ROOT.TGTextButton(self.fButtonFrame, "Hide")
        disp = ROOT.TPyDispatcher(self.HideClicked)
        self.fHideButton.Connect("Clicked()", "TPyDispatcher", disp, "Dispatch()")
        self._dispatchers.append(disp)
        self.fButtonFrame.AddFrame(self.fHideButton)

        self.fDecompButton = ROOT.TGCheckButton(self.fButtonFrame, "Show decomposition")
        disp = ROOT.TPyDispatcher(self.DecompClicked)
        self.fDecompButton.Connect("Clicked()", "TPyDispatcher", disp, "Dispatch()")
        self._dispatchers.append(disp)
        self.fButtonFrame.AddFrame(
            self.fDecompButton, ROOT.TGLayoutHints(ROOT.kLHintsLeft, 10, 0, 0, 0)
        )

        self.fMainFrame.AddFrame(
            self.fButtonFrame, ROOT.TGLayoutHints(ROOT.kLHintsExpandX, 10, 5, 10, 10)
        )

        ## Fit info ##
        self.fFitInfo = ROOT.TGTab(self.fMainFrame)
        self.fMainFrame.AddFrame(
            self.fFitInfo, ROOT.TGLayoutHints(ROOT.kLHintsExpandX | ROOT.kLHintsExpandY)
        )

        optionsFrame = self.fFitInfo.AddTab("Options")
        fitFrame = self.fFitInfo.AddTab("Fit")
        # listFrame = self.fFitInfo.AddTab("Peak list")

        self.fOptionsText = ROOT.TGTextView(optionsFrame, 400, 500)
        optionsFrame.AddFrame(
            self.fOptionsText,
            ROOT.TGLayoutHints(ROOT.kLHintsExpandX | ROOT.kLHintsExpandY),
        )

        self.fFitText = ROOT.TGTextView(fitFrame, 400, 500)
        fitFrame.AddFrame(
            self.fFitText, ROOT.TGLayoutHints(ROOT.kLHintsExpandX | ROOT.kLHintsExpandY)
        )

        self.fMainFrame.SetWindowName("Fit")
        self.fMainFrame.MapSubwindows()
        self.fMainFrame.Resize(self.fMainFrame.GetDefaultSize())

    # Note that deleting the window on close, and recreating it when needed,
    # causes a BadWindow (invalid Window parameter) error from time to time
    # (the exact conditions are not completely understood).
    # In addition, just hiding the window has the advantage that all settings
    # (text entries, checkboxes, ...) are automatically remembered.
    def Hide(self):
        if self.fVisible:
            self.fMainFrame.UnmapWindow()
            self.fVisible = False

    def Show(self):
        if not self.fVisible:
            self.fMainFrame.MapWindow()
            self.fVisible = True

    # FIXME: This should *really* take advantage of signals and slots...
    def FitClicked(self):
        if self.fFitHandler:
            self.fFitHandler()

    def ClearClicked(self):
        if self.fClearHandler:
            self.fClearHandler()

    def ResetClicked(self):
        if self.fResetHandler:
            self.fResetHandler()

    def HideClicked(self):
        self.Hide()

    def DecompClicked(self):
        if self.fDecompHandler:
            self.fDecompHandler(bool(self.fDecompButton.IsOn()))

    def SetDecomp(self, stat):
        if stat:
            self.fDecompButton.SetState(ROOT.kButtonDown)
        else:
            self.fDecompButton.SetState(ROOT.kButtonUp)

    def SetData(self, text):
        if not text:
            self.fFitText.Clear()
        else:
            self.fFitText.LoadBuffer(text)

    def SetOptions(self, text):
        if not text:
            self.fOptionsText.Clear()
        else:
            self.fOptionsText.LoadBuffer(text)



#!/usr/bin/env python3

import versioneer
from setuptools import setup
from pathlib import Path
import glob

manpages = glob.glob("doc/guide/*.1")

KEYWORDS = """\
analysis
data-analysis
gamma-spectroscopy
nuclear-physics
nuclear-spectrum-analysis
physics
python
root
root-cern
spectroscopy
"""
CLASSIFIERS = """\
Development Status :: 5 - Production/Stable
Environment :: Console
Environment :: X11 Applications
Intended Audience :: Information Technology
Intended Audience :: Science/Research
License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)
Natural Language :: English
Operating System :: MacOS
Operating System :: POSIX
Operating System :: POSIX :: Linux
Programming Language :: C
Programming Language :: C++
Programming Language :: Python
Programming Language :: Python :: 3
Programming Language :: Python :: 3.6
Programming Language :: Python :: 3.7
Programming Language :: Python :: 3.8
Topic :: Scientific/Engineering
Topic :: Scientific/Engineering :: Information Analysis
Topic :: Scientific/Engineering :: Physics
Topic :: Scientific/Engineering :: Visualization
"""

setup(
    name="hdtv",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="HDTV - Nuclear Spectrum Analysis Tool",
    long_description=(Path(__file__).parent / "README.md").read_text(),
    long_description_content_type="text/markdown",
    url="https://github.com/janmayer/hdtv",
    maintainer="Jan Mayer",
    maintainer_email="jan.mayer@ikp.uni-koeln.de",
    license="GPL",
    classifiers=CLASSIFIERS.strip().split("\n"),
    keywords=KEYWORDS.strip().replace("\n", " "),
    install_requires=[
        "scipy",
        "matplotlib",
        "numpy",
        "ipython",
        "prompt_toolkit>=3.0.0",
        "traitlets",
        "uncertainties",
    ],
    extras_require={
        "dev": ["docutils"],
        "test": ["pytest", "pytest-cov"],
    },
    entry_points={
        "console_scripts": [
            "hdtv=hdtv.app:run",
        ]
    },
    packages=[
        "hdtv",
        "hdtv.database",
        "hdtv.backgroundmodels",
        "hdtv.efficiency",
        "hdtv.peakmodels",
        "hdtv.plugins",
        "hdtv.rootext",
    ],
    package_data={
        "hdtv": ["share/*"],
        "hdtv.rootext": [
            "calibration/*",
            "display/*",
            "fit/*",
            "mfile-root/*",
            "mfile-root/*/*",
            "mfile-root/*/*/*",
        ],
    },
    data_files=[
        ("share/man/man1", manpages),
        ("share/zsh/site-functions", ["data/completions/_hdtv"]),
        ("share/bash-completion/completions", ["data/completions/hdtv"]),
        ("share/applications", ["data/hdtv.desktop"]),
    ],
)



# HDTV - A ROOT-based spectrum analysis software
#  Copyright (C) 2006-2009  The HDTV development team (see file AUTHORS)
#
# This file is part of HDTV.
#
# HDTV is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# HDTV is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with HDTV; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import re
import os
import sys
import filecmp

import pytest

from tests.helpers.utils import redirect_stdout, hdtvcmd
from tests.helpers.fixtures import temp_file

from hdtv.util import monkey_patch_ui

monkey_patch_ui()

import hdtv.cmdline
import hdtv.options
import hdtv.session

import __main__

try:
    __main__.spectra = hdtv.session.Session()
except RuntimeError:
    pass


from hdtv.plugins.specInterface import spec_interface
import hdtv.plugins.calInterface
from hdtv.plugins.fitInterface import fit_interface
import hdtv.plugins.peakfinder
import hdtv.plugins.fitmap

spectra = __main__.spectra

testspectrumfile = "osiris_bg.spc"
testspectrum = os.path.join(os.path.curdir, "tests", "share", testspectrumfile)


@pytest.fixture(autouse=True)
def prepare():
    fit_interface.ResetFitterParameters()
    hdtv.options.Set("table", "classic")
    hdtv.options.Set("uncertainties", "short")
    for _ in range(3):
        spec_interface.LoadSpectra(testspectrum)
    spectra.ActivateObject("0")
    yield
    spectra.Clear()


def test_cmd_cal_pos_set():
    f, ferr = hdtvcmd("calibration position set 1 2")
    assert ferr == ""
    assert "Calibrated spectrum with id 0" in f


def test_cmd_cal_pos_set_3():
    f, ferr = hdtvcmd("calibration position set 1 2 0.0001")
    assert ferr == ""
    assert "Calibrated spectrum with id 0" in f


def test_cmd_cal_pos_unset():
    hdtvcmd("calibration position set 1 2")
    f, ferr = hdtvcmd("calibration position unset")
    assert ferr == ""
    assert "Unsetting calibration of spectrum with id 0" in f


def test_cmd_cal_pos_copy():
    hdtvcmd("calibration position set 1 2 -s 0")
    f, ferr = hdtvcmd("calibration position copy 0 1")
    assert ferr == ""
    assert "Calibrated spectrum with id 1" in f


def test_cmd_cal_pos_enter():
    f, ferr = hdtvcmd("calibration position enter 1543 1173.228(3) 1747 1332.492(4)")
    assert ferr == ""
    assert "Calibrated spectrum with id 0" in f
    assert "Chi" in f


def test_cmd_cal_pos_enter_underdefined():
    f, ferr = hdtvcmd(
        "calibration position enter -d 2 1543 1173.228(3) 1747 1332.492(4)"
    )
    assert "You must specify at least" in ferr


def test_cmd_cal_pos_enter_deg2():
    f, ferr = hdtvcmd(
        "calibration position enter -d 2 1543 1173.228(3) 1747 1332.492(4) 3428 2614.5"
    )
    assert ferr == ""
    assert "Calibrated spectrum with id 0" in f


def test_cmd_cal_pos_enter_table():
    f, ferr = hdtvcmd("calibration position enter -t 1543 1173.228(3) 1747 1332.492(4)")
    assert ferr == ""
    assert "Residual" in f


def test_cmd_cal_pos_assign():
    hdtvcmd("fit peakfind -a -t 0.005")
    f, ferr = hdtvcmd("cal position assign 3.0 1173.228(3) 5.0 1332.492(4)")
    assert ferr == ""
    assert "Calibrated spectrum with id 0" in f


def test_cmd_cal_pos_list():
    hdtvcmd("cal pos set 1 2")
    f, ferr = hdtvcmd("cal position list")
    assert ferr == ""
    assert testspectrumfile + ": 1.0   2.0" in f


@pytest.mark.parametrize("callistfile", ["tests/share/callist.cal"])
def test_cmd_cal_pos_list_read(callistfile):
    f, ferr = hdtvcmd(f"cal position list read {callistfile}")
    assert ferr == ""
    assert "Calibrated spectrum with id 0" in f


@pytest.mark.parametrize("callistfile", ["tests/share/callist.cal"])
def test_cmd_cal_pos_list_clear(callistfile):
    hdtvcmd(f"cal position list read {callistfile}")
    f, ferr = hdtvcmd("cal position list clear")
    assert ferr == ""
    assert "Unsetting calibration of spectrum with id 0" in f


@pytest.mark.parametrize("callistfile", ["tests/share/callist.cal"])
def test_cmd_cal_pos_list_write(callistfile, temp_file):
    hdtvcmd(f"cal position list read {callistfile}")
    f, ferr = hdtvcmd(f"cal position list write -F {temp_file}")
    assert ferr == ""
    assert f == ""
    assert filecmp.cmp(callistfile, temp_file)


def test_cmd_cal_pos_recalibrate():
    hdtvcmd("fit peakfind -a -t 0.005")
    hdtvcmd("fit position assign 3.0 1173.228(3) 5.0 1332.492(4)")
    f, ferr = hdtvcmd("calibration position recalibrate")
    assert ferr == ""
    assert "Calibrated spectrum with id 0" in f
    assert "Chi" in f


@pytest.mark.parametrize("calfile", ["tests/share/osiris_bg.cal"])
def test_cmd_cal_pos_read(calfile):
    f, ferr = hdtvcmd(f"calibration position read {calfile}")
    assert ferr == ""
    assert "Calibrated spectrum with id 0" in f


@pytest.mark.parametrize("nuclide", ["Co-60", "Cs-137"])
def test_cmd_nuclide(nuclide):
    f, ferr = hdtvcmd(f"nuclide {nuclide}")
    assert ferr == ""
    assert "Nuclide:" in f
    assert "energy" in f
    assert "intensity" in f


@pytest.mark.parametrize(
    "parameters, function",
    [
        ("0.1,0.2,0.3,0.4,0.5", "pow"),
        ("0.1,0.2,0.3,0.4,0.5", "exp"),
        ("0.1,0.2,0.3,0.4,0.5", "poly"),
        ("0.1,0.2,0.3,0.4,0.5", "wiedenhoever"),
        ("0.1,0.2,0.3,0.4,0.5", "wunder"),
    ],
)
def test_cmd_cal_eff_set(parameters, function):
    f, ferr = hdtvcmd(f"calibration efficiency set -p {parameters} {function}")
    assert ferr == ""
    assert f == ""


@pytest.mark.parametrize("args", ["", "0", "all", "0 1", "0,1"])
def test_cmd_cal_eff_list(args):
    f, ferr = hdtvcmd(f"calibration efficiency list {args}")
    assert ferr == ""
    assert "Parameter" in f


@pytest.mark.parametrize(
    "parfile, efffunction", [("tests/share/osiris_bg.par", "wunder")]
)
def test_cmd_cal_eff_read_par(parfile, efffunction):
    hdtvcmd(f"calibration efficiency set {efffunction}")
    f, ferr = hdtvcmd(f"calibration efficiency read parameter {parfile}")
    assert ferr == ""
    assert f == ""


@pytest.mark.parametrize(
    "parfile, covfile, efffunction",
    [("tests/share/osiris_bg.par", "tests/share/osiris_bg.cov", "wunder")],
)
def test_cmd_cal_eff_read_cov(parfile, covfile, efffunction):
    hdtvcmd(f"calibration efficiency set {efffunction}")
    hdtvcmd(f"calibration efficiency read parameter {parfile}")
    f, ferr = hdtvcmd(f"calibration efficiency read covariance {covfile}")
    assert ferr == ""
    assert f == ""


@pytest.mark.skipif(
    sys.version_info < (3, 0), reason="Floating Point differences in py2"
)
@pytest.mark.parametrize(
    "parfile, covfile, efffunction",
    [("tests/share/osiris_bg.par", "tests/share/osiris_bg.cov", "wunder")],
)
def test_cmd_cal_eff_write_par(parfile, covfile, efffunction, temp_file):
    hdtvcmd(f"calibration efficiency set {efffunction}")
    hdtvcmd(f"calibration efficiency read parameter {parfile}")
    hdtvcmd(f"calibration efficiency read covariance {covfile}")
    f, ferr = hdtvcmd(f"cal efficiency write parameter {temp_file}")
    assert ferr == ""
    assert f == ""
    assert filecmp.cmp(parfile, temp_file)


@pytest.mark.skipif(
    sys.version_info < (3, 0), reason="Floating Point differences in py2"
)
@pytest.mark.parametrize(
    "parfile, covfile, efffunction",
    [("tests/share/osiris_bg.par", "tests/share/osiris_bg.cov", "wunder")],
)
def test_cmd_cal_eff_write_cov(parfile, covfile, efffunction, temp_file):
    hdtvcmd(f"calibration efficiency set {efffunction}")
    hdtvcmd(f"calibration efficiency read parameter {parfile}")
    hdtvcmd(f"calibration efficiency read covariance {covfile}")
    f, ferr = hdtvcmd(f"cal efficiency write covariance {temp_file}")
    assert ferr == ""
    assert f == ""
    assert filecmp.cmp(covfile, temp_file)


@pytest.mark.skip(reason="Sample spectrum not sufficient for test?")
def test_cmd_cal_pos_nuclide():
    raise NotImplementedError


@pytest.mark.skip(reason="Sample spectrum not sufficient for test?")
def test_cmd_cal_eff_fit():
    raise NotImplementedError


@pytest.mark.skip(reason="Hard to test")
def test_cmd_cal_eff_plot():
    raise NotImplementedError



#!/usr/bin/env python3


import hdtv.efficiency as eff

e = eff.WunderEff()

e.load("eff_wunder.par", "eff_wunder.cov")

for i in range(1, 10000, 100):
    print("Wunder-Efficiency @", i, " keV = ", e(i))
